/* 
 * Copyright (c) 2017, CoolR Group Inc 
 * All Rights Reserved. 
 * 
 * File: bleDrv.c 
 * 
 * Description:  APIs for accessing SoC bluetooth peripheral
 * 
 * Author: Cupola 
 * 
 */

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "counter.h"
#include "nrf.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "nrf_sdh.h"
#include "nrf_ble_gatt.h"
#include "nrf_sdh_ble.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "boards.h"
#include "amt.h"
#include "ble_advdata.h"
//#include "softdevice_handler.h"
#include "app_timer.h"
#include "bsp.h"
#include "bleDrv.h"
#include "ble_advertising.h"
#include "ble_db_discovery.h"
#include "beacon.h"
#include "task.h"
#include "SEGGER_RTT.h"
#include "nrf_fstorage.h"
#include "dataManager.h"
#include "mpu9250.h"
#include "motionActivity.h"
#include "isl29034.h"
#include "config.h"
#include "flash.h"
#include "rtc.h"
#include "coolerBoard.h"
//upgrade logic
#include "nrf_dfu_settings.h"
#include "nrf_delay.h"
#include "deepSleep.h"
#include "timer.h"
#include "crc32.h"
#include "ambientLight.h"

#include "nrf_cli.h"
#include "nrf_cli_rtt.h"
#include "nrf_cli_uart.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "bsp_btn_ble.h"


#include "camera.h"
#include "fpgaUpgrade.h"

#define DEVICE_NAME_NEW "CSS-429100867095"

#define SCAN_ADV_LED                    BSP_BOARD_LED_0
#define READY_LED                       BSP_BOARD_LED_1
#define PROGRESS_LED                    BSP_BOARD_LED_2
#define DONE_LED                        BSP_BOARD_LED_3

#define BOARD_TESTER_BUTTON             BSP_BUTTON_2                                    /**< Button to press at beginning of the test to indicate that this board is connected to the PC and takes input from it via the UART. */
#define BOARD_DUMMY_BUTTON              BSP_BUTTON_3                                    /**< Button to press at beginning of the test to indicate that this board is standalone (automatic behavior). */
#define BUTTON_DETECTION_DELAY          APP_TIMER_TICKS(50)                             /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */

#define APP_BLE_CONN_CFG_TAG            1                                               /**< A tag that refers to the BLE stack configuration. */
#define APP_BLE_OBSERVER_PRIO           3                                               /**< Application's BLE observer priority. You shouldn't need to modify this value. */
#define L2CAP_HDR_LEN                   4                                               /**< L2CAP header length. */




#define CONN_INTERVAL_DEFAULT           (uint16_t)(MSEC_TO_UNITS(7.5, UNIT_1_25_MS))    /**< Default connection interval used at connection establishment by central side. */

#define CONN_INTERVAL_MIN               (uint16_t)(MSEC_TO_UNITS(7.5, UNIT_1_25_MS))    /**< Minimum acceptable connection interval, in 1.25 ms units. */
#define CONN_INTERVAL_MAX               (uint16_t)(MSEC_TO_UNITS(500, UNIT_1_25_MS))    /**< Maximum acceptable connection interval, in 1.25 ms units. */

#define IS_SRVC_CHANGED_CHARACT_PRESENT     0                                       /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/

#if (NRF_SD_BLE_API_VERSION == 3)
//#define NRF_BLE_MAX_MTU_SIZE                GATT_MTU_SIZE_DEFAULT                   /**< MTU size used in the softdevice enabling and to reply to a BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST event. */
#define NRF_BLE_MAX_MTU_SIZE                BLE_GATT_ATT_MTU_DEFAULT                   /**< MTU size used in the softdevice enabling and to reply to a BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST event. */
#endif

#define CENTRAL_LINK_COUNT                  0                                       /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT               1                                       /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

#define APP_FEATURE_NOT_SUPPORTED           BLE_GATT_STATUS_ATTERR_APP_BEGIN + 2    /**< Reply when unsupported features are requested. */

// User-modifiable configuration parameters.
//      The following values shall be altered when doing power profiling.

#define APP_CFG_NON_CONN_ADV_TIMEOUT  30                                            /**< Time for which the device must be advertising in non-connectable mode (in seconds). */
#define APP_CFG_CHAR_NOTIF_TIMEOUT    5000                                          /**< Time for which the device must continue to send notifications once connected to central (in milli seconds). */
#define APP_CFG_ADV_DATA_LEN          31                                            /**< Required length of the complete advertisement packet. This should be atleast 8 in order to accommodate flag field and other mandatory fields and one byte of manufacturer specific data. */
#define APP_CFG_CONNECTION_INTERVAL   20                                            /**< Connection interval used by the central (in milli seconds). This application will be sending one notification per connection interval. A repeating timer will be started with timeout value equal to this value and one notification will be sent everytime this timer expires. */
#define APP_CFG_CHAR_LEN              20                                            /**< Size of the characteristic value being notified (in bytes). */

// Fixed configuration parameters:
//      The following parameters are not meant to be changed while using this application for power
//      profiling.

#define NOTIF_BUTTON_ID               0                                             /**< Button used for initializing the application in connectable mode. */
#define NON_CONN_ADV_BUTTON_ID        1                                             /**< Button used for initializing the application in non-connectable mode. */

#define ADV_DEVICE_NAME                   	"SmartTag2"                           /**< Name of device. Will be included in the advertising data. */
#define	PASSWORD							"password"

#define APP_TIMER_PRESCALER           0                                             /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE       4                                             /**< Size of timer operation queues. */

#define CHAR_NOTIF_TIMEOUT_IN_TKS     APP_TIMER_TICKS(APP_CFG_CHAR_NOTIF_TIMEOUT,\
                                                      APP_TIMER_PRESCALER)          /**< Time for which the device must continue to send notifications once connected to central (in ticks). */

#define NON_CONNECTABLE_ADV_INTERVAL  MSEC_TO_UNITS(100, UNIT_0_625_MS)             /**< The advertising interval for non-connectable advertisement (100 ms). This value can vary between 100ms to 10.24s). */
#define CONNECTABLE_ADV_TIMEOUT       0                                            /**< Time for which the device must be advertising in connectable mode (in seconds). */

#define SLAVE_LATENCY                 0                                             /**< Slave latency. */
#define CONN_SUP_TIMEOUT              MSEC_TO_UNITS(4000, UNIT_10_MS)               /**< Connection supervisory timeout (4 seconds). */

#define ADV_ENCODED_AD_TYPE_LEN       1                                             /**< Length of encoded ad type in advertisement data. */
#define ADV_ENCODED_AD_TYPE_LEN_LEN   1                                             /**< Length of the 'length field' of each ad type in advertisement data. */
#define ADV_FLAGS_LEN                 1                                             /**< Length of flags field that will be placed in advertisement data. */
#define ADV_ENCODED_FLAGS_LEN         (ADV_ENCODED_AD_TYPE_LEN +       \
                                       ADV_ENCODED_AD_TYPE_LEN_LEN +   \
                                       ADV_FLAGS_LEN)                               /**< Length of flags field in advertisement packet. (1 byte for encoded ad type plus 1 byte for length of flags plus the length of the flags itself). */
#define ADV_ENCODED_COMPANY_ID_LEN    2                                             /**< Length of the encoded Company Identifier in the Manufacturer Specific Data part of the advertisement data. */
#define ADV_ADDL_MANUF_DATA_LEN       (APP_CFG_ADV_DATA_LEN -                \
                                       (                                     \
                                           ADV_ENCODED_FLAGS_LEN +           \
                                           (                                 \
                                               ADV_ENCODED_AD_TYPE_LEN +     \
                                               ADV_ENCODED_AD_TYPE_LEN_LEN + \
                                               ADV_ENCODED_COMPANY_ID_LEN    \
                                           )                                 \
                                       )                                     \
                                      )                                             /**< Length of Manufacturer Specific Data field that will be placed on the air during advertisement. This is computed based on the value of APP_CFG_ADV_DATA_LEN (required advertisement data length). */

#define NOTIFY_BEFORE_DISCONNECT	(1)
#define DISCONNECT					(2)
#define DISCONNECT_NOTIFICATION_TIMEOUT	100
#if APP_CFG_ADV_DATA_LEN > BLE_GAP_ADV_MAX_SIZE
    #error "The required advertisement data size (APP_CFG_ADV_DATA_LEN) is greater than the value allowed by stack (BLE_GAP_ADV_MAX_SIZE). Reduce the value of APP_CFG_ADV_DATA_LEN and recompile."
#endif

// Check whether the maximum characteristic length + opcode length (1) + handle length (2) is not
// greater than default MTU size.
//#if (APP_CFG_CHAR_LEN + 1 + 2) > BLE_L2CAP_MTU_DEF
#if (APP_CFG_CHAR_LEN + 1 + 2) > BLE_L2CAP_MTU_MIN
    #error "The APP_CFG_CHAR_LEN is too large for the maximum MTU size."
#endif

#if ADV_ADDL_MANUF_DATA_LEN < 1
    #error "The required length of additional manufacturer specific data computed based on the user configured values is computed to be less than 1. Consider increasing the value of APP_CFG_ADV_DATA_LEN."
#endif



typedef enum
{
    NOT_SELECTED = 0x00,
    BOARD_TESTER,
    BOARD_DUMMY,
} board_role_t;

// Name to use for advertising and connection.
static char const m_target_periph_name[] = DEVICE_NAM_NEW;

typedef struct
{
    uint16_t        att_mtu;                    /**< GATT ATT MTU, in bytes. */
    uint16_t        conn_interval;              /**< Connection interval expressed in units of 1.25 ms. */
    ble_gap_phys_t  phys;                       /**< Preferred PHYs. */
    bool            data_len_ext_enabled;       /**< Data length extension status. */
    bool            conn_evt_len_ext_enabled;   /**< Connection event length extension status. */
} test_params_t;

/**@brief Variable length data encapsulation in terms of length and pointer to data. */
typedef struct
{
    uint8_t  * p_data;      /**< Pointer to data. */
    uint16_t   data_len;    /**< Length of data. */
} data_t;


NRF_BLE_GATT_DEF(m_gatt);                   /**< GATT module instance. */
BLE_DB_DISCOVERY_DEF(m_ble_db_discovery);   /**< DB discovery module instance. */
NRF_CLI_UART_DEF(cli_uart, 0, 64, 16);
NRF_CLI_RTT_DEF(cli_rtt);

static nrf_ble_amtc_t     m_amtc;
static nrf_ble_amts_t     m_amts;

NRF_CLI_DEF(m_cli_uart, "throughput example:~$ ", &cli_uart.transport, '\r', 4);
NRF_CLI_DEF(m_cli_rtt,  "throughput example:~$ ", &cli_rtt.transport,  '\n', 4);


static bool volatile m_run_test;
static bool volatile m_notif_enabled;
static bool volatile m_mtu_exchanged;
static bool volatile m_data_length_updated;
static bool volatile m_phy_updated;
static bool volatile m_conn_interval_configured;

static board_role_t volatile m_board_role  = NOT_SELECTED;
static uint8_t m_gap_role     = BLE_GAP_ROLE_INVALID;       /**< BLE role for this connection, see @ref BLE_GAP_ROLES */

// Test parameters.
// Settings like ATT MTU size are set only once on the dummy board.
// Make sure that defaults are sensible.
static test_params_t m_test_params =
{
    .att_mtu                  = NRF_SDH_BLE_GATT_MAX_MTU_SIZE,
    .conn_interval            = CONN_INTERVAL_DEFAULT,
    .data_len_ext_enabled     = true,
    .conn_evt_len_ext_enabled = true,
    // Only symmetric PHYs are supported.
#if defined(S140)
    .phys.tx_phys             = BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS | BLE_GAP_PHY_CODED,
    .phys.rx_phys             = BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS | BLE_GAP_PHY_CODED,
#else
    .phys.tx_phys             = BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS,
    .phys.rx_phys             = BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS,
#endif
};


// Scan parameters requested for scanning and connection.
static ble_gap_scan_params_t const m_scan_param =
{
    .active         = 0x00,
    .interval       = SCAN_INTERVAL,
    .window         = SCAN_WINDOW,
    .use_whitelist  = 0x00,
    .adv_dir_report = 0x00,
    .timeout        = 0x0000, // No timeout.
};

// Connection parameters requested for connection.
static ble_gap_conn_params_t m_conn_param =
{
    .min_conn_interval = CONN_INTERVAL_MIN,   // Minimum connection interval.
    .max_conn_interval = CONN_INTERVAL_MAX,   // Maximum connection interval.
    .slave_latency     = SLAVE_LATENCY,       // Slave latency.
    .conn_sup_timeout  = CONN_SUP_TIMEOUT     // Supervisory timeout.
};



uint8_t commandBuffer[CMD_MAX_SIZE];
uint16_t size = 0;
app_timer_t disconnect_timer;
app_timer_t *disconnectTimerId = &disconnect_timer; //put rhis and ble connection handle in global data
notificationRspStruct_t response = {};

//

uint8_t readBuffer[100] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23};
/**@brief 128-bit UUID base List. */
static const ble_uuid128_t m_base_uuid128 =
	{
		//45b06612-2416-11e7-93ae-92361f002671 Service UUID
		//45b06b26-2416-11e7-93ae-92361f002671 Char UUID
		// The 128 bit UUID base is 45b0XXXX-2416-11e7-93ae-92361f002671
		{
			0x71, 0x26, 0x00, 0x1F, 0x36, 0x92, 0xAE, 0x93, 0xE7, 0x11, 0x16, 0x24, 0x00, 0x00, 0xB0, 0x45}};

// added for testing
static ble_uuid_t m_adv_uuids[] = {
	{LOCAL_SERVICE_UUID, BLE_UUID_TYPE_BLE},
}; //!< Universally unique service identifiers. */


static ble_gap_adv_params_t     m_adv_params;                                       /**< Parameters to be passed to the stack when starting advertising. */
static uint8_t                  m_char_value[APP_CFG_CHAR_LEN];                     /**< Value of the characteristic that will be sent as a notification to the central. */
static uint8_t                  m_addl_adv_manuf_data[ADV_ADDL_MANUF_DATA_LEN];     /**< Value of the additional manufacturer specific data that will be placed in air (initialized to all zeros). */
static ble_gatts_char_handles_t m_char_handles;                                     /**< Handles of local characteristic (as provided by the BLE stack).*/
uint16_t                 m_conn_handle = BLE_CONN_HANDLE_INVALID;            /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection).*/
static uint16_t                 m_service_handle;                                   /**< Handle of local service (as provided by the BLE stack).*/
static bool                     m_is_notifying_enabled = false;                     /**< Variable to indicate whether the notification is enabled by the peer.*/
APP_TIMER_DEF(m_conn_int_timer_id);                                                 /**< Connection interval timer. */
APP_TIMER_DEF(m_notif_timer_id);                                                    /**< Notification timer. */
MPU_Data_t Data;

static void test_terminate(void);


static void instructions_print(void)
{
    printf("Type 'config' to change the configuration parameters.");
    printf("You can use the TAB key to autocomplete your input.");
    printf("Type 'run' when you are ready to run the test.");
}



/**@brief Function for starting advertising. */
static void advertising_start(void)
{
    ble_gap_adv_params_t const adv_params =
    {
        .type        = BLE_GAP_ADV_TYPE_ADV_IND,
        .p_peer_addr = NULL,
        .fp          = BLE_GAP_ADV_FP_ANY,
        .interval    = ADV_INTERVAL,
        .timeout     = 0,
    };

    printf("Starting advertising.");

    bsp_board_led_on(SCAN_ADV_LED);

    ret_code_t err_code = sd_ble_gap_adv_start(&adv_params, APP_BLE_CONN_CFG_TAG);
    APP_ERROR_CHECK(err_code);
}



/***************************************************************
 * Task handler to set the Reset password
 **************************************************************/
uint32_t setResetPassTaskHandler(uint32_t id, void *data, uint32_t len)
{
	uint8_t isNested = 0;
	uint32_t ret = NRF_SUCCESS;
//	bleServiceCommand_t *cmdData = data;
#if 0
	// Enter critical section
//	sd_nvic_critical_region_enter(&isNested);
	NRF_UICR_Type uicrBkp;
	uint32_t *uicrBuf = &uicrBkp;
	uint32_t *uicrReg = NRF_UICR;

	memcpy(&uicrBkp, NRF_UICR, sizeof(NRF_UICR_Type));

	memcpy((uint8_t *)uicrBuf +RESET_PASS_ADDR_OFFSET,cmdData->commandData.auth.password,RESET_PASSWORD_SIZE);

	printF("written data = %08x\n",uicrBkp.CUSTOMER[0]);

#if 0
	/* MWU Enable -- This is workaround suggested by Aryan on the Nordic Developer Zone */
	NRF_MWU->REGIONENSET
		= ((MWU_REGIONENSET_RGN0WA_Set << MWU_REGIONENSET_RGN0WA_Pos)
				| (MWU_REGIONENSET_PRGN0WA_Set << MWU_REGIONENSET_PRGN0WA_Pos));
	//#endif // 0 or 1
#endif // 0 or 1
	// Erase the UICR register

	NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Een << NVMC_CONFIG_WEN_Pos);
	while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}
	printF("Erase enabled\n");
	NRF_NVMC->ERASEUICR = (NVMC_ERASEUICR_ERASEUICR_Erase << NVMC_ERASEUICR_ERASEUICR_Pos);
	while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}

	/* Turn on flash write enable and wait until the NVMC is ready: */
	NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Wen << NVMC_CONFIG_WEN_Pos);
	while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}

	printF("waiting for write complete\n");
	/* write to the register */
	//  memcpy(&(NRF_UICR->CUSTOMER[0]), &(uicrBkp.CUSTOMER[0]), 8);

	for(int i = 1; i < sizeof(NRF_UICR_Type) / 4; i++)
	{
		uicrReg[i] = uicrBuf[i];
		while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}
	}

	//  NRF_UICR->CUSTOMER[0] = 0x1234;
	//while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}

	printF("Write complete\n");

	/* Turn off flash write enable and wait until the NVMC is ready: */
	NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Ren << NVMC_CONFIG_WEN_Pos);
	while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}

	printF("Write complete\n");
	printF("nested = %d\n", isNested);
//	if(isNested != true)
//		sd_nvic_critical_region_exit(isNested);
	print(LL_INFO, "Password of reset-password is set.\n");
#else
	oneTimeConfiguration_Structure *oneTimeConfig = (oneTimeConfiguration_Structure *)ONE_TIME_CONFIG_FLASH_ADDR;
	oneTimeConfiguration_Structure bkp;

	memcpy(&bkp, oneTimeConfig, sizeof(oneTimeConfiguration_Structure));

	memcpy(&bkp.resetPassword, data, sizeof(bkp.resetPassword));

	flashErasePage(ONE_TIME_CONFIG_FLASH_ADDR / FLASH_PAGE_SIZE, 1);

	nrf_delay_ms(100);

	flashWrite((uint32_t *)&bkp, (uint32_t *)oneTimeConfig, sizeof(bkp) / WORD_SIZE);

	return ret;

#endif // 0 or 1
}

/**@brief AMT server event handler. */
static void amts_evt_handler(nrf_ble_amts_evt_t evt)
{
    ret_code_t err_code;

    switch (evt.evt_type)
    {
        case NRF_BLE_AMTS_EVT_NOTIF_ENABLED:
        {
            NRF_LOG_INFO("Notifications enabled.");

            bsp_board_led_on(READY_LED);
            m_notif_enabled = true;

            if (m_board_role != BOARD_TESTER)
            {   
                return;
            }

            if (m_test_params.conn_interval != CONN_INTERVAL_DEFAULT)
            {
                NRF_LOG_DEBUG("Updating connection parameters..");
                m_conn_param.min_conn_interval = m_test_params.conn_interval;
                m_conn_param.max_conn_interval = m_test_params.conn_interval;
                err_code = sd_ble_gap_conn_param_update(m_conn_handle, &m_conn_param);

                if (err_code != NRF_SUCCESS)
                {   
                    NRF_LOG_ERROR("sd_ble_gap_conn_param_update() failed: 0x%x.", err_code);
                }
            }
            else
            {
                m_conn_interval_configured = true;
            }
        } break;

        case NRF_BLE_AMTS_EVT_NOTIF_DISABLED:
        {
            NRF_LOG_INFO("Notifications disabled.");
            bsp_board_led_off(READY_LED);
        } break;

        case NRF_BLE_AMTS_EVT_TRANSFER_1KB:
        {
            NRF_LOG_INFO("Sent %u KBytes", (evt.bytes_transfered_cnt / 1024));
            bsp_board_led_invert(PROGRESS_LED);
        } break;

        case NRF_BLE_AMTS_EVT_TRANSFER_FINISHED:
        {
            counter_stop();

            bsp_board_led_off(PROGRESS_LED);
            bsp_board_led_on(DONE_LED);

            uint32_t time_ms      = counter_get();
            uint32_t bit_count    = (evt.bytes_transfered_cnt * 8);
            float throughput_kbps = ((bit_count / (time_ms / 1000.f)) / 1000.f);

            NRF_LOG_INFO("Done.");
            NRF_LOG_INFO("=============================");
            NRF_LOG_INFO("Time: %u.%.2u seconds elapsed.", (time_ms / 1000), (time_ms % 1000));
            NRF_LOG_INFO("Throughput: " NRF_LOG_FLOAT_MARKER " Kbps.",
                         NRF_LOG_FLOAT(throughput_kbps));
            NRF_LOG_INFO("=============================");
            NRF_LOG_INFO("Sent %u bytes of ATT payload.", evt.bytes_transfered_cnt);
            NRF_LOG_INFO("Retrieving amount of bytes received from peer...");

            err_code = nrf_ble_amtc_rcb_read(&m_amtc);
            if (err_code != NRF_SUCCESS)
            {
                NRF_LOG_ERROR("nrf_ble_amtc_rcb_read() failed: 0x%x.", err_code);
                test_terminate();
            }
        } break;
    }
}




/**@brief Function to start scanning. */
static void scan_start(void)
{
    printf("Starting scan.");

    bsp_board_led_on(SCAN_ADV_LED);

    ret_code_t err_code = sd_ble_gap_scan_start(&m_scan_param);
    APP_ERROR_CHECK(err_code);
}




static void test_terminate(void)
{
    m_run_test                 = false;
    m_notif_enabled            = false;
    m_mtu_exchanged            = false;
    m_data_length_updated      = false;
    m_phy_updated              = false;
    m_conn_interval_configured = false;

    if (m_conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        printf("Disconnecting...");

        ret_code_t err_code;
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);

        if (err_code != NRF_SUCCESS)
        {
            NRF_LOG_ERROR("sd_ble_gap_disconnect() failed: 0x%0x.", err_code);
        }
    }
    else
    {
        if (m_board_role == BOARD_DUMMY)
        {
            if (m_gap_role == BLE_GAP_ROLE_PERIPH)
            {
                advertising_start();
            }
            else
            {
                scan_start();
            }
        }
    }
}




char const * phy_str(ble_gap_phys_t phys)
{
    static char const * str[] =
    {
        "1 Mbps",
        "2 Mbps",
        "Coded",
        "Unknown"
    };

    switch (phys.tx_phys)
    {
        case BLE_GAP_PHY_1MBPS:
            return str[0];

        case BLE_GAP_PHY_2MBPS:
        case BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS:
        case BLE_GAP_PHY_2MBPS | BLE_GAP_PHY_1MBPS | BLE_GAP_PHY_CODED:
            return str[1];

        case BLE_GAP_PHY_CODED:
            return str[2];

        default:
            return str[3];
    }
}
/***************************************************************
 * Initialization function for task
 **************************************************************/
static void disconnectTimerHandler(void *pEvent)
{
	uint32_t res;
	if (m_conn_handle != BLE_CONN_HANDLE_INVALID)
	{
		res = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		m_conn_handle = BLE_CONN_HANDLE_INVALID;
		print(LL_INFO, "res = %d\n", res);
	}
}

static void disconnectTimerStart(void)
{
	uint32_t ret = NRF_SUCCESS;
	//ret = app_timer_start(disconnectTimerId, APP_TIMER_TICKS(globalData.disconnectTimeout, 0), NULL);
	ret = app_timer_start(disconnectTimerId, APP_TIMER_TICKS(globalData.disconnectTimeout), NULL);
	if (ret != NRF_SUCCESS)
	{
		print(LL_ERROR, "app_timer_start returned error %d \n", ret);
	}
}

static void disconnectTimerStop(void)
{
	uint32_t ret = NRF_SUCCESS;
	ret = app_timer_stop(disconnectTimerId);
	if (ret != NRF_SUCCESS)
	{
		print(LL_ERROR, "app_timer_stop returned error %d \n", ret);
	}
}

static void disconnectTimerInit(void)
{
	uint32_t ret = NRF_SUCCESS;
	ret = app_timer_create(&disconnectTimerId, APP_TIMER_MODE_SINGLE_SHOT, disconnectTimerHandler);
	if (ret != NRF_SUCCESS)
	{
		print(LL_ERROR, "app_timer_create returned error %d \n", ret);
	}
}

static void disconnectTimerRestart(void)
{
	uint32_t ret = NRF_SUCCESS;
	// Stop the already running timer
	disconnectTimerStop();

	//Start the timer again
	disconnectTimerStart();
}

uint32_t bleDisconnectTaskHandler(uint32_t Id, void *data, uint32_t len)
{
	uint32_t err_code = 0;
	uint8_t isNested;

	print(LL_INFO, "Data = %08x\n",data);
	if(*(uint8_t*)data == NOTIFY_BEFORE_DISCONNECT)
	{
		uint8_t disconnect = DISCONNECT;
		taskEnqueue(globalData.bleDisconnectTaskId, &disconnect, sizeof(disconnect), TASK_PRIORITY_LOW, DISCONNECT_NOTIFICATION_TIMEOUT);
		notifyData(RESPONSE_TYPE, (uint8_t *)&response, sizeof(response));
		return 0;
	}
	else if(*(uint8_t*)data == DISCONNECT)
	{
		err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		m_conn_handle = BLE_CONN_HANDLE_INVALID;
		print(LL_INFO, "error code = %d\n", err_code);
		return 0;
	}
/*	else if(globalData.doorStatus == DOOR_OPENED || globalData.isDeviceMoving == 1) // any activity such as door/movement, then wait till the activity is done
	{
		taskEnqueue(globalData.bleDisconnectTaskId, NULL, 0, TASK_PRIORITY_LOW, DISCONNECT_NOTIFICATION_TIMEOUT);
		advertisingStart();
		return 0;
	}
*/	else
	{
		print(LL_INFO, "disconnect handler called\n");
		advertisingDataInit();

		globalData.bleLinkAuthenticationStatus = false;
		globalData.authenticationUsingPasswordResetStatus = false;
		if (globalData.factoryResetTriggerFlag == true)
		{
			globalData.factoryResetTriggerFlag = false;

			//corrupt the global data
			globalData.magicNumber[0] = 0;

			//trigger save config and system reboot
			globalData.configurationSettingFlag = true;
			globalData.rebootTriggerFlag = true;
		}

		// For upgrade trigger
		if (globalData.upgradeTriggerFlag == true)
		{
			globalData.upgradeTriggerFlag = false;

			print(LL_INFO, "bootloader flag = %d\n", s_dfu_settings.enter_buttonless_dfu);

			s_dfu_settings.enter_buttonless_dfu = true;
			updateDeviceInfo();
			strcpy(s_dfu_settings.deviceName, globalData.deviceName);
			memcpy(s_dfu_settings.macAddr, globalData.macAddr, MAC_ADDR_SIZE);

			s_dfu_settings.watchdogTimeout = globalData.watchdogTimeout;

			//for next time trigger upgrade

			print(LL_INFO, "bootloader flag = %d\n", s_dfu_settings.enter_buttonless_dfu);

			flashErasePage(BOOTLOADER_SETTING_FLASH_ADDR / FLASH_PAGE_SIZE, 1);

			print(LL_INFO, "flash erase error code = %d\n", err_code);

			s_dfu_settings.crc = 0; //nrf_dfu_settings_calculate_crc();

			err_code = flashWrite((uint32_t *)&s_dfu_settings, (uint32_t *)BOOTLOADER_SETTING_FLASH_ADDR, sizeof(s_dfu_settings) / WORD_SIZE);
			if (err_code == NRF_ERROR_BUSY)
			{
				print(LL_INFO, "flash write error code = %d\n", err_code);
			}

			print(LL_INFO, "bootloader flag = %d\n", ((nrf_dfu_settings_t *)(BOOTLOADER_SETTING_FLASH_ADDR))->enter_buttonless_dfu);
			//			nrf_delay_ms(5000);
			globalData.rebootTriggerFlag = true;
			globalData.configurationSettingFlag = true;
		}

		if (globalData.configurationSettingFlag == true)
		{
			if (globalData.deepSleepTriggerFlag == true)
			{
				writethevalue();
				resetDoorStatus();
				globalData.standByFlag = true;
			}
			globalData.configurationSettingFlag = false;

			// Log the debug information to flash
			if (logDebugInfotoFlash())
			{
				print(LL_ERROR, "Failed to log the debug info to flash\n");
			}


			saveCurrentConfig();
		}

		if (globalData.deepSleepTriggerFlag == true)
		{

			globalData.deepSleepTriggerFlag = false;
			deepSleepHandler();
		}

		if (globalData.rebootTriggerFlag == true)
		{
			globalData.rebootTriggerFlag = false;

			print(LL_DEBUG, "Rebooting\n");
#if WATCHDOG_TIMER
			//sd_nvic_critical_region_enter(&isNested);
			//		while(true);
			NVIC_SystemReset();
#else
			NVIC_SystemReset();
#endif //WATCHDOG_TIMER
		}

		if (globalData.resetPasswordWriteFlag == true)
		{
			oneTimeConfiguration_Structure *oneTimeConfig = (oneTimeConfiguration_Structure *)ONE_TIME_CONFIG_FLASH_ADDR;
			oneTimeConfiguration_Structure bkp;

			memcpy(&bkp, oneTimeConfig, sizeof(oneTimeConfiguration_Structure));

			memcpy(&bkp.resetPassword, globalData.tempResetPassword, sizeof(bkp.resetPassword));

			flashErasePage(ONE_TIME_CONFIG_FLASH_ADDR / FLASH_PAGE_SIZE, 1);

			nrf_delay_ms(100);

			flashWrite((uint32_t *)&bkp, (uint32_t *)oneTimeConfig, sizeof(bkp) / WORD_SIZE);
			memset(globalData.tempResetPassword, 0x00, RESET_PASSWORD_SIZE);
			globalData.resetPasswordWriteFlag = false;
		}

		advertisingStart();
		return 0;
	}
}

uint16_t serviceAuthenticate(uint8_t *password)
{
	oneTimeConfiguration_Structure *oneTimeConfig = (oneTimeConfiguration_Structure *)ONE_TIME_CONFIG_FLASH_ADDR;
	if (strcmp((char *)password, (char *)globalData.bleAuthenticationPassword) == 0)
	{
		globalData.bleLinkAuthenticationStatus = 1;
		print(LL_ERROR, "BLE Authentication success..\n");
		return 1;
	}
	else if (strcmp((char *)password, (char *)oneTimeConfig->resetPassword) == 0)
	{
		globalData.authenticationUsingPasswordResetStatus = 1;
		print(LL_DEBUG, "Set-password Authentication successful\n");
		return 1;
	}
	else
	{
		globalData.bleLinkAuthenticationStatus = 0;
		globalData.authenticationUsingPasswordResetStatus = 0;

		print(LL_ERROR, "BLE Authentication Failed %x %x %x %x %x %x\n", globalData.bleAuthenticationPassword[0], globalData.bleAuthenticationPassword[1], globalData.bleAuthenticationPassword[2], globalData.bleAuthenticationPassword[3], globalData.bleAuthenticationPassword[4], globalData.bleAuthenticationPassword[5]);

		print(LL_ERROR, "Authentication Failed.... %s\n", (uint8_t *)oneTimeConfig->resetPassword);
		print(LL_ERROR, "pass %02x\n", oneTimeConfig->resetPassword[0]);
		print(LL_ERROR, "pass %02x\n", oneTimeConfig->resetPassword[1]);
		print(LL_ERROR, "pass %02x\n", oneTimeConfig->resetPassword[1]);
		print(LL_ERROR, "pass addr  %08x\n", oneTimeConfig->resetPassword);
	}

	return 0;
}

uint16_t bleAttributeWriteCallBack(ble_uuid_t uuid, uint16_t handle, uint8_t *data, uint16_t *len)
{
	uint32_t res;
	uint32_t evtLocation = 0;
	bleServiceCommand_t *commandReq;
	uint8_t flag;
	uint8_t commandProcessingRequired;
	uint8_t cfgGetLen = 0;
	configGetStruct_t configGetRsp = {};
	uint8_t eventDataPayload[MAX_NOTIFY_LEN] = {};
	uint16_t eventDataPayloadLen = 0;
	bool responseNeeded = 1;
	uint16_t ret;

	//memset((uint8_t*)&notification,0,sizeof(notificationStruct_t));

	if (handle != m_char_handles.value_handle) /* Characteristic descriptor uuid or handle */
	{
		// Return with invalid access - need to find the approriate error value for return, but for now we can return success
		return BLE_GATT_STATUS_SUCCESS;
	}

	if (*len < MIN_COMMAND_SIZE)
	{
		// Return with invalid command status - need to find the approriate error value for return, but for now we can return success
		return BLE_GATT_STATUS_SUCCESS;
	}

	flag = data[0];
	commandProcessingRequired = 0;

	print(LL_INFO, "Packet format = %d\n", (flag & PKT_POS_MASK));
	switch (flag & PKT_POS_MASK)
	{
	case SINGLE_CMD_PKT:
		size = 0;
		commandReq = (bleServiceCommand_t *)((uint8_t *)data + 1);
		size += (*(uint8_t *)len) - 1; //only for passing length
		commandProcessingRequired = 1;
		break;
	case START_CMD_PKT:
		size = 0;
		memcpy(commandBuffer + size, (uint8_t *)(data + 1), (*(uint8_t *)len) - 1);
		size += (*(uint8_t *)len) - 1;
		break;
	case END_CMD_PKT:
		commandProcessingRequired = 1;
		commandReq = (bleServiceCommand_t *)commandBuffer;
	case MID_CMD_PKT:
		memcpy(commandBuffer + size, (uint8_t *)(data + 1), (*(uint8_t *)len) - 1);
		size += (*(uint8_t *)len) - 1;
		break;

	default:
		print(LL_ERROR, "Invalid packet format\n");
		return BLE_GATT_STATUS_SUCCESS;
	}

	if (!commandProcessingRequired)
	{
		return BLE_GATT_STATUS_SUCCESS;
	}

	// Assign the response command ID and initialize the response with success response
	response.commandId = commandReq->commandType;
	response.rspCode = RSP_SUCCESS;
	
#ifdef DEBUG_INFO
	if (commandReq->commandType >= AUTHENTICATE && commandReq->commandType < NUM_OF_COMMANDS)
	{
		if ((commandReq->commandType != AUTHENTICATE) &&
			(commandReq->commandType != GET_ALL_CONFIG_DATA) &&
			(commandReq->commandType != CONFIGURATION_GET))
		{
			logDebugInfo(BLUETOOTH_COMMAND_LOG, (uint8_t *)commandReq, size);
		}
	}
#endif // DEBUG_INFO

	switch (commandReq->commandType)
	{

#if	CAMERA

		case IMAGE_CAPTURE_TRIGGER:
			responseNeeded = 0;						// Responses will be handled independently by camera modules
			triggerCamReadyImgCapture();
		break;

		case IMAGE_DOWNLOAD_TRIGGER:
			responseNeeded = 0;						// Responses will be handled independently by camera modules
			if(size == sizeof(uint16_t))					// If the command does not have sequence number in its arguments, then download latest image
			{
				triggerUnreadImageDownload();
			}
			else if(size == (sizeof(uint16_t) + sizeof(uint32_t)) )		// If the command has sequence number in its arguments, then download image with the seq num provided
			{
				triggerImageDownload(commandReq->commandData.imageDownload.seqNo);
			}
			else if(size == (sizeof(uint16_t) + 2 * sizeof(uint32_t)) )	// If the command has sequence number and packet number in its arguments, then resume downloading of the image with the seq num provided from the packet number provided
			{
				resumeImageDownload(commandReq->commandData.imageResumeDownload.seqNo,commandReq->commandData.imageResumeDownload.pktNo);
			}
			else
			{
				// Send error response ??? need to check as to response from the generic frame work or camera protocols
				responseNeeded = 1;
				response.rspCode = RSP_INVALID_PARAM;
			}
		break;

		case IMAGE_STATISTICS_TABLE:
			responseNeeded = 0;				// Responses will be handled independently by camera modules
			notifyImageStatistics();
		break;

		case IMAGE_DELETE:
			responseNeeded = 0;                             // Responses will be handled independently by camera modules
                        if(size == sizeof(uint16_t))    	        // If the command does not have sequence number in its arguments, then delete all images
                        {
                                deleteAllImages();           
                        }
                        else                                            // If the command has sequence number in its arguments, then delete image with the seq num provided
                        {
                                deleteImage(commandReq->commandData.imageDelete.seqNo);
                        }
		break;
		case FPGA_UPGRADE_TRIGGER:
		{
			responseNeeded = 0;
			fpgaUpgradeInit(commandReq->commandData.fpgaUpgradeTrigger.size,commandReq->commandData.fpgaUpgradeTrigger.crc);
		}
		break;
		case FPGA_UPGRADE_FIRMWARE_DATA:
		{
			responseNeeded = 0;  
			if(globalData.fpgaUpgradeInfo.fpgaUpgradeState != FPGA_DOWNLOAD_BINARY)
			{
				response.rspCode = RSP_UNKNOWN_FAILURE;
			}
			taskEnqueue(globalData.fpgaUpgradeTaskId, (void *)&(commandReq->commandData.fpgaUpgradeData), 0, TASK_PRIORITY_HIGH, sizeof(commandReq->commandData.fpgaUpgradeData));
		}
		break;

#endif // CAMERA


	// Authenticate the connection
	case AUTHENTICATE:
	{
		// Authenticate the link from the password provided
		uint16_t ret = serviceAuthenticate(commandReq->commandData.auth.password);

		// Create the response accordingly
		if (globalData.bleLinkAuthenticationStatus == 0 && globalData.authenticationUsingPasswordResetStatus == 0)
			response.rspCode = RSP_AUTHENTICATION_FAILURE;
	}
	break;
	// Get single or multiple (not implemented yet) selected configuration(s)
	case CONFIGURATION_GET:
	{
		// if the configuration to be obtained is link security status then add an exception for access
		if (commandReq->commandData.config.id != BLE_LINK_AUTHENTICATION_STATUS && globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		// Update the configuration ID in the notification
		configGetRsp.id = commandReq->commandData.config.id;

		// Send any optional data if any
		memcpy(configGetRsp.data,commandReq->commandData.config.data,size - sizeof(commandReq->commandType) - sizeof(commandReq->commandData.config.id) );
		// Get the configuration and send it over notification
		ret = configGetData(configGetRsp.id, configGetRsp.data, &(configGetRsp.length));

		// If the configuration is present then notify the data else send an error response
		if (ret == NRF_SUCCESS)
		{
			// No response is needed as the configuration will be sent via notification
			responseNeeded = 0;
			notifyData(CONFIGURATION_TYPE, (uint8_t *)&configGetRsp, configGetRsp.length + sizeof(configGetRsp) - sizeof(configGetRsp.data));
		}
		else
		{
			// Configuration get failure, as the configuration ID is not present
			response.rspCode = RSP_DOESNOT_EXIST;
		}
	}
	break;

	case CONFIGURATION_SET:
	{
		// If the connection is not secured then send authentication required response
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		ret = configSetData(commandReq->commandData.config.id, commandReq->commandData.config.data, size - sizeof(commandReq->commandType) - sizeof(commandReq->commandData.config.id));
		// If configuration set function returns success then respond success via notification
		if (ret == NRF_SUCCESS)
		{
			//globalData.configurationSettingFlag = true;
		}
		else
		{
			// Configuration set failure, as the configuration ID is not present
			response.rspCode = ret;
		}
	}
	break;
	case DATA_SYNC_ACK:
	{
		// If the connection is not secured then send authentication required response
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		// Acknowledge the event number
		res = eventDataSyncAck(commandReq->commandData.dataSyncAck.eventNo);

		// If the operation failed then return error else do not notify anything to save time and power during data transfer
		if (res != DM_SUCCESS)
			response.rspCode = RSP_UNKNOWN_FAILURE;
		else
			responseNeeded = 0;
	}
	break;
	case REBOOT:
		// If the connection is not secured then send authentication required response
		if ((globalData.bleLinkAuthenticationStatus != 1) && (globalData.authenticationUsingPasswordResetStatus != 1))
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		responseNeeded = 0;
		// Set the reboot flag
		print(LL_INFO, "Rebooting...\n");
		globalData.rebootTriggerFlag = true;
		//uint8_t isNested;
		// Enter critical section
		//sd_nvic_critical_region_enter(&isNested);
		//while(1);
		break;
	case UPGRADE:
		// If the connection is not secured then send authentication required response
		if ((globalData.bleLinkAuthenticationStatus != 1) && (globalData.authenticationUsingPasswordResetStatus != 1))
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		print(LL_INFO, "Upgrade request receive \n");
		globalData.upgradeTriggerFlag = true;
		// disconnect ble and do trigger in disconnect task
		responseNeeded = 0;
		break;
	case FACTORY_RESET:
		// If the connection is not secured then send authentication required response
		if ((globalData.bleLinkAuthenticationStatus != 1) && (globalData.authenticationUsingPasswordResetStatus != 1))
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		//keep in factory reset task
		//res = factoryReset();
		globalData.factoryResetTriggerFlag = true;
		responseNeeded = 0;
		break;

	case GET_EVENT_DATA:

		// If the connection is not secured then send authentication required response
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		// Check for the start event number and end event number present in the data
		if((size) > (sizeof((*(bleServiceCommand_t *)data).commandType)))
		{
			// Check argument validity
			if (((size) - (sizeof((*(bleServiceCommand_t *)data).commandType))) <= sizeof(uint16_t))
			{
				if ((uint16_t)globalData.flashStartEventNo < commandReq->commandData.eventDataInfo.startEventNumber)
				{
					dataManagerEvtSync.eventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))) * (globalData.flashStartEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))))) + commandReq->commandData.eventDataInfo.startEventNumber;
					print(LL_DEBUG, "Flash start event number is less than start event number: %d\n", dataManagerEvtSync.eventNumber);
				}
				else
				{
					dataManagerEvtSync.eventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))) * (globalData.dataManagerInfo.latestEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))))) + commandReq->commandData.eventDataInfo.startEventNumber;
					print(LL_DEBUG, "Flash start event number is greater than start event number: %d\n", dataManagerEvtSync.eventNumber);					
				}
				dataManagerEvtSync.endEventNumber = globalData.dataManagerInfo.latestEventNo;

				print(LL_DEBUG, "1 argument in the command detected\n");
			}
			else
			{
				if ((uint16_t)globalData.flashStartEventNo < commandReq->commandData.eventDataInfo.startEventNumber)
				{
					print(LL_DEBUG, "Flash start event number is less than start event number\n");
					dataManagerEvtSync.eventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))) * (globalData.flashStartEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))))) + commandReq->commandData.eventDataInfo.startEventNumber;
				}
				else
				{
					print(LL_DEBUG, "Flash start event number is greater than start event nymber\n");
					dataManagerEvtSync.eventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))) * (globalData.dataManagerInfo.latestEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.startEventNumber))))) + commandReq->commandData.eventDataInfo.startEventNumber;
				}

				if ((uint16_t)globalData.dataManagerInfo.latestEventNo > commandReq->commandData.eventDataInfo.eventInfo.endEventNumber)
				{
					print(LL_DEBUG, "Flash latest event number is greater than start event number\n");
					dataManagerEvtSync.endEventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.eventInfo.endEventNumber))) * (globalData.dataManagerInfo.latestEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.eventInfo.endEventNumber))))) + commandReq->commandData.eventDataInfo.eventInfo.endEventNumber;
				}
				else
				{
					dataManagerEvtSync.endEventNumber = ((1 << (8 * sizeof(commandReq->commandData.eventDataInfo.eventInfo.endEventNumber))) * (globalData.flashStartEventNo / (1 << (8 * sizeof(commandReq->commandData.eventDataInfo.eventInfo.endEventNumber))))) + commandReq->commandData.eventDataInfo.eventInfo.endEventNumber;
				}

				print(LL_DEBUG, "2 arguments in the command detected\n");
			}
		}
		else
		{
			dataManagerEvtSync.eventNumber = globalData.dataManagerInfo.noofEventSynced + 1;
			dataManagerEvtSync.endEventNumber = globalData.dataManagerInfo.latestEventNo;
			print(LL_DEBUG, "No argument in the command\n");
		}

		evtLocation = locateEventNo(dataManagerEvtSync.eventNumber);
		if ((evtLocation != DM_EVT_IN_RAM) && (evtLocation != DM_EVT_IN_FLASH))
		{
			response.rspCode = RSP_INVALID_PARAM;
			print(LL_ERROR, "Invalid event number is provided\n");
			break;
		}
		evtLocation = locateEventNo(dataManagerEvtSync.endEventNumber);
		if ((evtLocation != DM_EVT_IN_RAM) && (evtLocation != DM_EVT_IN_FLASH))
		{
			response.rspCode = RSP_INVALID_PARAM;
			print(LL_ERROR, "Invalid event number is provided\n");
			break;
		}
		// Check for the event number validity
		// if((dataManagerEvtSync.eventNumber < (uint16_t)globalData.flashStartEventNo) || 
		// 	(dataManagerEvtSync.endEventNumber > (uint16_t)globalData.dataManagerInfo.latestEventNo))
		// {
		// 	response.rspCode = RSP_INVALID_PARAM;
		// 	print(LL_ERROR, "Invalid event number is provided\n");
		// 	break;
		// }

		print(LL_INFO, "Event data to be provided from %d to %d\n", dataManagerEvtSync.eventNumber, dataManagerEvtSync.endEventNumber);
		if(dataManagerEvtSync.eventNumber > dataManagerEvtSync.endEventNumber)
		{
			response.rspCode = RSP_INVALID_PARAM;
			print(LL_ERROR,"Start event number is greater than end event number.\n");
			break;
		}
		dataManagerEvtSync.firstTimeFlag = true;

		// Enqueue the task event for event data notify task to send the event data one by one
		taskEnqueue(globalData.eventDataNotifyTaskId, (uint8_t *)&dataManagerEvtSync, sizeof(dataManagerEvtSync), TASK_PRIORITY_HIGH, 0);

		// Response is not needed as the actual data will be notified
		responseNeeded = 0;

		break;
	case SET_CURRENT_EVENT_NUMBER:
		// If the connection is not secured then send authentication required response

		
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		if ((uint16_t)globalData.flashStartEventNo < commandReq->commandData.eventNumber)
		{
			globalData.dataManagerInfo.noofEventSynced = ((1 << (8 * sizeof(commandReq->commandData.eventNumber))) * (globalData.flashStartEventNo / (1 << (8 * sizeof(commandReq->commandData.eventNumber))))) + commandReq->commandData.eventNumber;
			print(LL_DEBUG, "Flash start event number is less than start event number: %d\n", globalData.dataManagerInfo.noofEventSynced);
		}
		else
		{
			globalData.dataManagerInfo.noofEventSynced = ((1 << (8 * sizeof(commandReq->commandData.eventNumber))) * (globalData.dataManagerInfo.latestEventNo / (1 << (8 * sizeof(commandReq->commandData.eventNumber))))) + commandReq->commandData.eventNumber;
			print(LL_DEBUG, "Flash start event number is greater than start event number: %d\n", globalData.dataManagerInfo.noofEventSynced);					
		}

		evtLocation = locateEventNo(globalData.dataManagerInfo.noofEventSynced);
		if ((evtLocation != DM_EVT_IN_RAM) && (evtLocation != DM_EVT_IN_FLASH))
		{
			response.rspCode = RSP_INVALID_PARAM;
			print(LL_ERROR, "Invalid event number is provided\n");
			break;
		}

		print(LL_INFO, "BLE Event data %d\n", globalData.dataManagerInfo.noofEventSynced);
		// if (commandReq->commandData.eventNumber != 0)
		// {
		// 	globalData.dataManagerInfo.noofEventSynced = ((1 << (8 * sizeof(commandReq->commandData.eventNumber))) * (globalData.flashStartEventNo / (1 << (8 * sizeof(commandReq->commandData.eventNumber))))) + commandReq->commandData.eventNumber;
		// }
		// else
		// {
		// 	globalData.dataManagerInfo.noofEventSynced = globalData.flashStartEventNo;
		// }
		globalData.dataManagerInfo.noofEventSynced = globalData.dataManagerInfo.noofEventSynced - 1;

		break;

	case GET_ALL_CONFIG_DATA:
		// If the connection is not secured then send authentication required response
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		print(LL_INFO, "Notifying the configuration data\n");
		taskEnqueue(globalData.configurationNotifyTaskId, NULL, 0, TASK_PRIORITY_HIGH, 0);

		// Response is not needed as the actual data will be notified
		responseNeeded = 0;
		break;

#if DEEP_SLEEP_ENABLED
	case ACTIVATE_DEEP_SLEEP:
		{
			// if (globalData.bleLinkAuthenticationStatus != 1)
			// {
			// 	response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			// 	break;
			// }

			/* Turn off flash write enable and wait until the NVMC is ready: */

			//		uint32_t* pointer   =   (uint32_t*)0X1000109C;
			/*		uint32_t* pointer   =   (uint32_t*)DEEP_SLEEP_FLAG;
			memcpy(&flag, (uint32_t *)DEEP_SLEEP_FLAG,  4);
			print(LL_INFO,"....................1\n");
			NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Wen << NVMC_CONFIG_WEN_Pos);
			while (NRF_NVMC->READY == NVMC_READY_READY_Busy){

			print(LL_INFO,"....................2\n");
			}
			if(*pointer == 0xFFFFFFFFFF)
			{
				NRF_NVMC->CONFIG = (NVMC_CONFIG_WEN_Wen << NVMC_CONFIG_WEN_Pos);
				while (NRF_NVMC->READY == NVMC_READY_READY_Busy){}
				*pointer = 0x1234;
				printF("*pointer =%x\n",*pointer );	
			}
			*/
			uint8_t lux = 0;
			isl29034PowerDown(0);
			nrf_delay_ms(200);
			if ((lux = readAmbientLight()) <= AL_DEEP_SLEEP_THRESHOLD)
			{
				globalData.configurationSettingFlag = true;
				globalData.deepSleepTriggerFlag = true;
				responseNeeded = 0;
				break;
			}
			print(LL_INFO, "not perfectly covered...%u\n", lux);
			response.rspCode = RSP_OPERATION_FAILED;
		}
		break;
#endif // DEEP_SLEEP_ENABLED
#if STANDBY_MODE

	case STAND_BY_TRIGGER:
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		if (commandReq->commandData.standby.standbyFlag == 0)
		{
			print(LL_INFO, "StandBy Disabled...\n");
			globalData.standByFlag = false;
		}
		else
		{
			print(LL_INFO, "StandBy Enabled...\n");
			globalData.standByFlag = true;
		}

		//globalData.configurationSettingFlag = true;
		break;

#endif // STANDBY_MODE
	case RESET_DOOR:
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		globalData.resetDoorFlag = 1;
		break;

	case ERASE_ALL_EVENT_DATA:
		print(LL_INFO, "Erasing all event data...\n");
		//dataManagerInit();
		globalData.dataManagerInfo.noofEventSynced = globalData.dataManagerInfo.latestEventNo;
		print(LL_INFO, "Erasing event data complete\n");
		break;

	case INSTALL_DEVICE:
		if (globalData.bleLinkAuthenticationStatus != 1)
		{
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		globalData.isMpuCalibrated = 0;
		taskEnqueue(globalData.installDeviceTaskId, NULL, 0, TASK_PRIORITY_HIGH, 0);
		break;

	case SET_RESET_PASSWORD:
	{
		if (globalData.authenticationUsingPasswordResetStatus != 1)
		{

			print(LL_INFO, "Setting the Password of reset-password is Failed.\n");
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}

		if((size) > (RESET_PASSWORD_STRING_SIZE +  sizeof((*(bleServiceCommand_t *)data).commandType)))
		{
			print(LL_INFO, "Invalid reset password length %d\n", size);
			response.rspCode = RSP_INVALID_PARAM;
			break;
		}
		uint8_t enteredPass[RESET_PASSWORD_SIZE] = {};
		strcpy((char *)enteredPass, (char *)commandReq->commandData.auth.password);
		int8_t i = 0;

		for (i = 0; enteredPass[i]; i++)

			if (!(enteredPass[i] >= '0' && enteredPass[i] <= '9'))
			{
				print(LL_INFO, "Entered password is not valid\n");
				response.rspCode = RSP_OPERATION_FAILED;
				break;
			}
		if (i != strlen((char *)enteredPass))
			break;

		globalData.resetPasswordWriteFlag = true;
		responseNeeded = 0;
		memcpy(globalData.tempResetPassword, enteredPass, RESET_PASSWORD_SIZE);
		// taskEnqueue(globalData.bleDisconnectTaskId, NULL, 0, TASK_PRIORITY_HIGH, 0);
		// taskEnqueue(globalData.setResetPassTaskId, commandReq->commandData.auth.password, MAX_PASSWORD_SIZE, TASK_PRIORITY_HIGH, 10);
	}
	break;

	case RESET_USER_PASSWORD:
	{

		if (globalData.authenticationUsingPasswordResetStatus != 1)
		{

			print(LL_INFO, "Password reset Authentication is required.\n");
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
		memcpy(globalData.bleAuthenticationPassword, globalData.bleAuthenticationDefaultPassword, MAX_PASSWORD_SIZE);

		print(LL_INFO, "User Password reset to Default Password.\n");
	}
	break;

	case DEBUG_INFO_AUTHENTICATE:
	{
		// Authenticate the link from the password provided
		if(commandReq->commandData.debugAuth == 0x11223344)
		{
			globalData.debugInfoAuthenticationStatus = true;
		}
		else
		{
			globalData.debugInfoAuthenticationStatus = false;
		}
	}
	break;

	case GET_DEBUG_INFO:
	{
		// Whether authentication is needed or not ???
		if (globalData.debugInfoAuthenticationStatus != true)
		{
			print(LL_ERROR, "debug Authentication is required\n");
			response.rspCode = RSP_AUTHENTICATION_REQUIRED;
			break;
		}
    	// If non bluetooth command is seen, notify error
		print(LL_INFO, "debug ID = %08X\n", commandReq->commandData.debugCommand.commandID);
		if ((commandReq->commandData.debugCommand.commandID < DEBUG_ID_BLUETOOTH_START) ||
			(commandReq->commandData.debugCommand.commandID > DEBUG_COMMAND_ID_MAX))
		{
			print(LL_ERROR, "Debug command ID out of range\n");
			response.rspCode = RSP_INVALID_PARAM;
			break;
		}
		taskEnqueue(globalData.debugInfoTaskId, (uint8_t *)&(commandReq->commandData.debugCommand), sizeof(commandReq->commandData.debugCommand), TASK_PRIORITY_HIGH, 0);
	}
	break;

	default:
		response.rspCode = RSP_DOESNOT_EXIST;
		break;
	}

	// Notify the response if response is needed by the central
	if (responseNeeded)
		notifyData(RESPONSE_TYPE, (uint8_t *)&response, sizeof(response));

	//	Reset the command buffer
	memset(commandBuffer, 0x00, sizeof(commandBuffer));
	size = 0;

	return BLE_GATT_STATUS_SUCCESS;
}

uint16_t bleAttributeReadCallBack(ble_uuid_t uuid, uint16_t handle, uint8_t *data, uint16_t *len)
{
	int i, ret;

	print(LL_INFO, "Read Call Back, UUID Type = %d UUID Value = %02X Handle = %d\n", uuid.type, uuid.uuid, handle);

	// If the connection is not secured then send authentication required response
	if (globalData.bleLinkAuthenticationStatus != 1)
	{
		*len = 0;
		return BLE_GATT_STATUS_SUCCESS; //	Insufficiant authentication error needs to be returned, this is temporary as the BLE app is not able to handle error
										//return BLE_GATT_STATUS_ATTERR_INSUF_AUTHENTICATION;
	}

	// Get the latest data
	ret = eventDataSync(data, len);
	if (ret == DM_ERR_SYNC_FAILED)
	{
		*len = 0;
		print(LL_DEBUG, "No Unsynced data remaining\n");
		return BLE_GATT_STATUS_SUCCESS;
	}

	print(LL_INFO, "Data received from Data Manager is ");
	for (i = 0; i < *len; i++)
		print(LL_PRINT, " %02X ", data[i]);
	print(LL_PRINT, "\n ", data[i]);

	return BLE_GATT_STATUS_SUCCESS;
}

uint8_t *hashFunction()
{
	static uint8_t temp[MAX_PASSWORD_SIZE] = {};
	int i = 0, j = 0;

	memcpy((char *)temp, (char *)globalData.macAddr, MAC_ADDR_SIZE);

#if MAC_ADDRESS_BASED_PASSWORD

	for (i = 0, j = 1; j <= MAC_ADDR_SIZE; i = i + 2, j = j + 2)
	{
		temp[i] = temp[j] ^ temp[i] ^ (temp[j] = temp[i]);
	}

	print(LL_INFO, "password is %x:%x:%x:%x:%x:%x \n", temp[0], temp[1], temp[2], temp[3], temp[4], temp[5]);

#else

	strncpy((char*)temp,(const char *)DEFAULT_PASSWORD,MAX_PASSWORD_SIZE - 1);

#endif // MAC_ADDRESS_BASED_PASSWORD

	return temp;
}

void bleAuthInit(void)
{
	oneTimeConfiguration_Structure* oneTimeConfig = (oneTimeConfiguration_Structure *)ONE_TIME_CONFIG_FLASH_ADDR;

	memset(globalData.bleAuthenticationPassword, 0, MAX_PASSWORD_SIZE);
	//	strcpy((char *)globalData.bleAuthenticationPassword,PASSWORD);	//initialize password
 	if((uint32_t) (* (uint32_t *) oneTimeConfig->defaultPassword) == 0xFFFFFFFF )													//Checking for the presense of default password
	{
		print(LL_INFO,"calculating default password...\n");

		strcpy((char *)globalData.bleAuthenticationDefaultPassword, (char *)hashFunction());									//creating Default password

		strcpy((char *)globalData.bleAuthenticationPassword, (char *)globalData.bleAuthenticationDefaultPassword);	//initialize password
	}
	else
	{
		
		print(LL_INFO,"Assigning default password...\n");
		//memcpy(globalData.bleAuthenticationDefaultPassword, oneTimeConfig->defaultPassword, TWO_WORDS);

		// Copy the default password from the one time configuration partition
		strncpy((char *) globalData.bleAuthenticationDefaultPassword, (char *) oneTimeConfig->defaultPassword, MAX_PASSWORD_SIZE);
		// Make sure that the string is NULL terminated
		globalData.bleAuthenticationDefaultPassword[MAX_PASSWORD_SIZE - 1] = '\0'; 

		// Copy the actual password comparision array with the default password
		strcpy((char *) globalData.bleAuthenticationPassword, (char *) globalData.bleAuthenticationDefaultPassword); 

	}
	// memset(globalData.deviceName,0 , MAX_DEVICE_NAME_SIZE);
	// strcpy((char *)globalData.deviceName,ADV_DEVICE_NAME);	//Initialize deviceName
	globalData.bleLinkAuthenticationStatus = 0;
}

uint16_t notifyData(uint8_t notificationID, uint8_t const *data, uint8_t len)
{
	uint32_t err_code = 0;
	uint16_t copyLen, index = 0;
	uint8_t notificationDataBuff[MAX_NOTIFY_LEN];
	uint8_t pktPos = 0xFF; // Intitalize invalid data to identify the begining of the loop

	if (notificationID >= MAX_NOTIFICATION_TYPES)
	{
		// Handle error if the notification id is wrong or exceeding the number of notification ids
		return NRF_ERROR_INVALID_PARAM;
	}

	// Break the complete notification data into one or more notification packets as the single notification size is limited to MAX_NOTIFY_PAYLOAD + 1
	while (len)
	{
		// If the length is greater than MAX_NOTIFY_PAYLOAD then break the data into chunks of MAX_NOTIFY_PAYLOAD
		if (len > MAX_NOTIFY_PAYLOAD)
		{
			// The amount of data present (copyLen) in the notification packet is MAX_NOTIFY_PAYLOAD as total length remaining is greater than MAX_NOTIFY_PAYLOAD
			copyLen = MAX_NOTIFY_PAYLOAD;

			// Subtract the copy length from the length remaining for calculating the notification packet size for the consecutive notification
			len -= copyLen;

			// Identify the packet position as start or middle and mark it accordingly
			if (pktPos == 0xFF)
				pktPos = START_CMD_PKT;
			else
				pktPos = MID_CMD_PKT;
		}
		else
		{
			// If the length is lesser than or equal to MAX_NOTIFY_PAYLOAD then notify the complete remaining notification data
			copyLen = len;

			// Subtract the copy length from the length remaining for exiting the loop
			len -= copyLen;

			// Identify the packet position as start or end and mark it accordingly
			if (pktPos == 0xFF)
				pktPos = SINGLE_CMD_PKT;
			else
				pktPos = END_CMD_PKT;
		}

		// Construct the notification header
		notificationDataBuff[0] = 0;																   // Reset to zero
		notificationDataBuff[0] |= pktPos;															   // Indicate the packet position
		notificationDataBuff[0] |= ((notificationID << NOTIFICATION_BIT_POS) & NOTIFICATION_BIT_MASK); // Indicate the notification type

		// Copy the payload from first byte as the 0th byte contains the notification header
		memcpy(notificationDataBuff + 1, data + index, copyLen);

		// Increment the index to keep track of the data position
		index += copyLen;

		// Create task event and enqueue for notification task
		err_code = taskEnqueue(globalData.notifyTaskId, notificationDataBuff, copyLen + 1, TASK_PRIORITY_LOW, 0);

		// Return any error
		if (err_code != NRF_SUCCESS)
			return err_code;
	}
	return NRF_SUCCESS;
}

uint32_t notifyTaskHandler(uint32_t Id, void *data, uint32_t len)
{
	uint32_t err_code;

	len = (uint16_t)len;

	// Send value if connected and notification is enabled
	if ((m_conn_handle != BLE_CONN_HANDLE_INVALID) && m_is_notifying_enabled)
	{
		ble_gatts_hvx_params_t hvx_params;

		memset(&hvx_params, 0, sizeof(hvx_params));

		hvx_params.handle = m_char_handles.value_handle;
		hvx_params.type = BLE_GATT_HVX_NOTIFICATION;
		hvx_params.offset = 0;
		hvx_params.p_len = (uint16_t *)&len;
		hvx_params.p_data = (uint8_t *)data;

		err_code = sd_ble_gatts_hvx(m_conn_handle, &hvx_params);
		if ((err_code != NRF_SUCCESS) &&
			(err_code != NRF_ERROR_INVALID_STATE) &&
			(err_code != BLE_ERROR_INVALID_ROLE) &&
			(err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING))
		{
			APP_ERROR_HANDLER(err_code);
		}
	}
        else
        {
                return NRF_ERROR_INVALID_STATE;
        }

	return NRF_SUCCESS;
}

/***********************************************************/ /**
 * @brief Function for the GAP initialization.
 *
 * @details This function shall be used to setup all the necessary GAP (Generic Access Profile)
 *          parameters of the device. It also sets the permissions.
 *
 **************************************************************/
void gapParamsInit(void)
{
#if 0
	uint32_t err_code;
	ble_gap_conn_params_t gap_conn_params;
	ble_gap_conn_sec_mode_t sec_mode;

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

	updateDeviceInfo();
    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)globalData.deviceName,
                                          strlen((char *)globalData.deviceName));
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_UNKNOWN);
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    // Set GAP Peripheral Preferred Connection Parameters (converting connection interval from
    // milliseconds to required unit of 1.25ms).
    gap_conn_params.min_conn_interval = (4 * APP_CFG_CONNECTION_INTERVAL) / 5;
    gap_conn_params.max_conn_interval = (4 * APP_CFG_CONNECTION_INTERVAL) / 5;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);

    #else
	ret_code_t              err_code;
    ble_gap_conn_sec_mode_t sec_mode;
	printf("gapParamsInit 00000000\n");

	nrf_delay_ms(250);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);
//	updateDeviceInfo();
	printf("gapParamsInit 11111111\n");
	nrf_delay_ms(250);
    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (uint8_t const *)DEVICE_NAME_NEW,
                                          strlen(DEVICE_NAME_NEW));
	printf("gapParamsInit 22222222\n");
	nrf_delay_ms(250);
    APP_ERROR_CHECK(err_code);
	err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_UNKNOWN);                                               
    APP_ERROR_CHECK(err_code); 

	// Set GAP Peripheral Preferred Connection Parameters (converting connection interval from
    // milliseconds to required unit of 1.25ms).
    m_conn_param.min_conn_interval = (4 * APP_CFG_CONNECTION_INTERVAL) / 5;
    m_conn_param.max_conn_interval = (4 * APP_CFG_CONNECTION_INTERVAL) / 5;
    m_conn_param.slave_latency     = SLAVE_LATENCY;
    m_conn_param.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&m_conn_param);
	printf("gapParamsInit 3333333\n");
    APP_ERROR_CHECK(err_code);
#endif 

}


/**@brief Function for handling events from the GATT library. */
static void gatt_evt_handler(nrf_ble_gatt_t * p_gatt, nrf_ble_gatt_evt_t const * p_evt)
{
    switch (p_evt->evt_id)
    {
        case NRF_BLE_GATT_EVT_ATT_MTU_UPDATED:
        {
            m_mtu_exchanged = true;
            printf("ATT MTU exchange completed. MTU set to %u bytes.",
                         p_evt->params.att_mtu_effective);
        } break;

        case NRF_BLE_GATT_EVT_DATA_LENGTH_UPDATED:
        {
            m_data_length_updated = true;
            printf("Data length updated to %u bytes.", p_evt->params.data_length);
        } break;
    }

    nrf_ble_amts_on_gatt_evt(&m_amts, p_evt);
}


/**@brief Function for initializing the GATT library. */
static void gatt_init(void)
{
    ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, gatt_evt_handler);
    APP_ERROR_CHECK(err_code);
}



/***********************************************************/ /**
 *@brief Function for initializing the connectable advertisement parameters.
 *
 * @details This function initializes the advertisement parameters to values that will put
 *          the application in connectable mode.
 *
 **************************************************************/

void connectableAdvInit(void)
{
	// Initialize advertising parameters (used when starting advertising).
	memset(&m_adv_params, 0, sizeof(m_adv_params));

	m_adv_params.type = BLE_GAP_ADV_TYPE_ADV_IND;
	m_adv_params.p_peer_addr = NULL; // Undirected advertisement
	m_adv_params.fp = BLE_GAP_ADV_FP_ANY;
	m_adv_params.interval = globalData.connectableAdvInterval;
	m_adv_params.timeout = CONNECTABLE_ADV_TIMEOUT;
}

static void on_adv_evt(ble_adv_evt_t ble_adv_evt)
{
}

/***********************************************************/ /**
 *@brief Function for initializing the Advertisement packet.
 *
 * @details This function initializes the data that is to be placed in an advertisement packet in
 *          both connectable and non-connectable modes.
 *
 **************************************************************/
void advertisingDataInit(void)
{
	uint32_t err_code;
	ble_advdata_t advdata;
	ble_advdata_t scanResponceData;
	ble_advdata_manuf_data_t manuf_data;
	ble_adv_modes_config_t options;
	uint8_t flags = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;
	sensorAdvDataStruct_t sensorAdvData;
	sensorScanRspDataStruct_t sensorScanRspData;
	manufactureData_t manufacturerData;
	ble_advdata_service_data_t serviceData;

	// Assert that these two values of the same.
	if (sizeof(flags) != ADV_FLAGS_LEN)
	{
		print(LL_ERROR, "Flag size is more\n");
		goto end;
	}

	//memset the advertising data and response data
	memset(&advdata, 0, sizeof(advdata));
	memset(&scanResponceData, 0, sizeof(scanResponceData));
	memset(&options, 0, sizeof(options));
	memset(&sensorAdvData, 0, sizeof(sensorAdvData));
	memset(&sensorScanRspData, 0, sizeof(sensorScanRspData));

#ifndef ADV_DATA_DEBUG
	//Set the Company identifier
	manuf_data.company_identifier = COMPANY_IDENTIFIER;

	//Set the manufacturing data to be advertised

	//Manufacture data should be papulated by bit packing informations.
	//Required to update the p_data with bit packed data???

	manufacturerData.doorStatus = globalData.doorStatus;
	manufacturerData.deviceMovementStatus = globalData.isDeviceMoving;
	manufacturerData.deviceFailure = globalData.sensorErrorFlag;

	if ((globalData.dataManagerInfo.latestEventNo) != (globalData.dataManagerInfo.noofEventSynced))
	{
		//print(LL_PRINT,"unsynced data is there latest = %d; unsynced = %d\n", globalData.dataManagerInfo.latestEventNo, globalData.dataManagerInfo.noofEventSynced);
		manufacturerData.dataAvailable = true;
	}
	else
	{
		//print(LL_PRINT,"unsynced data is not there latest = %d; unsynced = %d\n", globalData.dataManagerInfo.latestEventNo, globalData.dataManagerInfo.noofEventSynced);
		manufacturerData.dataAvailable = false;
	}
	manuf_data.data.p_data = &manufacturerData;
	manuf_data.data.size = sizeof(manufacturerData);
#else
	// copy all sensor data to adv data from global data
	sensorAdvData.doorStatus = globalData.doorStatus;
	sensorAdvData.deviceMovementStatus = globalData.isDeviceMoving;
	sensorAdvData.doorAngle = globalData.angle;
	sensorAdvData.deviceFailure = globalData.sensorErrorFlag;
	sensorAdvData.batteryCriticalStatus = globalData.batteryCriticalStatus;

	if ((globalData.dataManagerInfo.latestEventNo) != (globalData.dataManagerInfo.noofEventSynced))
	{
		//print(LL_PRINT,"unsynced data is there latest = %d; unsynced = %d\n", globalData.dataManagerInfo.latestEventNo, globalData.dataManagerInfo.noofEventSynced);
		sensorAdvData.dataAvailable = true;
	}
	else
	{
		//print(LL_PRINT,"unsynced data is not there latest = %d; unsynced = %d\n", globalData.dataManagerInfo.latestEventNo, globalData.dataManagerInfo.noofEventSynced);
		sensorAdvData.dataAvailable = false;
	}
	manuf_data.company_identifier = COMPANY_IDENTIFIER;
	manuf_data.data.p_data = (uint8_t *)&sensorAdvData;
	manuf_data.data.size = sizeof(sensorAdvData);
#endif
	advdata.flags = flags;
#ifndef ADV_DATA_DEBUG
	//update the scan response data
	scanResponceData.p_manuf_specific_data = &manuf_data;
#else
	//update health data in adv data
	advdata.p_manuf_specific_data = &manuf_data;
#endif

#ifndef ADV_DATA_DEBUG
	//Set the device name
	advdata.name_type = BLE_ADVDATA_FULL_NAME;
#else
	scanResponceData.name_type = BLE_ADVDATA_FULL_NAME;
#endif
	if (scanResponceData.name_type == BLE_ADVDATA_SHORT_NAME)
	{
		scanResponceData.short_name_len = BLE_ADVDATA_SHORT_NAME_LENGTH;
	}
#ifndef ADV_DATA_DEBUG
	advdata.include_appearance = true;
	advdata.uuids_complete.uuid_cnt = sizeof(m_adv_uuids) / sizeof(m_adv_uuids[0]);
	advdata.uuids_complete.p_uuids = m_adv_uuids;

	err_code = ble_advdata_set(&advdata, &scanResponceData);
#else
	// advdata.uuids_complete.uuid_cnt = sizeof(m_adv_uuids) / sizeof(m_adv_uuids[0]);
	// advdata.uuids_complete.p_uuids  = m_adv_uuids;

	err_code = sd_ble_uuid_vs_add(&m_base_uuid128, &m_adv_uuids[0].type);
	if (err_code != NRF_SUCCESS)
	{
		print(LL_INFO, "UUID Addition failed\n");
		goto end;
	}

	advdata.uuids_complete.uuid_cnt = sizeof(m_adv_uuids) / sizeof(m_adv_uuids[0]);
	advdata.uuids_complete.p_uuids = m_adv_uuids;

	sensorScanRspData.temperature = globalData.lastMeasuredTemperature;
	sensorScanRspData.humidity = globalData.lastMeasuredRelativeHumidity;
	sensorScanRspData.ambientLight = globalData.lastMeasuredAmbientLight;
	sensorScanRspData.doorTimeStamp = globalData.doorTimeStamp;

	serviceData.service_uuid = m_adv_uuids[0].uuid;
	serviceData.data.p_data = (uint8_t *)&sensorScanRspData;
	serviceData.data.size = sizeof(sensorScanRspData);
	scanResponceData.p_service_data_array = &serviceData;
	scanResponceData.service_data_count = 1;

	err_code = ble_advdata_set(&advdata, &scanResponceData);
#endif
	if (err_code != NRF_SUCCESS)
	{
		print(LL_ERROR, "setting the BLE advertisement data failed with error code %d\n", err_code);
	}
end:
	return;
}

/***********************************************************/ /**
 *@brief Function for adding the Characteristic.
 *
 * @details This function adds the characteristic to the local db.
 *
 * @param[in] uuid_type Type of service UUID assigned by the SoftDevice.
 *
 **************************************************************/
static void char_add(const uint8_t uuid_type)
{
	uint32_t err_code;
	ble_gatts_char_md_t char_md;
	ble_gatts_attr_md_t cccd_md;
	ble_gatts_attr_t attr_char_value;
	ble_uuid_t char_uuid;
	ble_gatts_attr_md_t attr_md;

	memset(&cccd_md, 0, sizeof(cccd_md));

	cccd_md.vloc = BLE_GATTS_VLOC_STACK;
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

	memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read = 1;
	char_md.char_props.notify = 1;
	char_md.char_props.write = 1;
	char_md.p_char_user_desc = NULL;
	char_md.p_char_pf = NULL;
	char_md.p_user_desc_md = NULL;
	char_md.p_cccd_md = &cccd_md;
	char_md.p_sccd_md = NULL;

	char_uuid.type = uuid_type;
	char_uuid.uuid = LOCAL_CHAR_UUID;

	memset(&attr_md, 0, sizeof(attr_md));

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
	//BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&attr_md.write_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

	attr_md.vloc = BLE_GATTS_VLOC_STACK;
	attr_md.rd_auth = 1;
	attr_md.wr_auth = 1;
	attr_md.vlen = 1;

	memset(&attr_char_value, 0, sizeof(attr_char_value));

	attr_char_value.p_uuid = &char_uuid;
	attr_char_value.p_attr_md = &attr_md;
	attr_char_value.init_len = NRF_BLE_MAX_MTU_SIZE;
	attr_char_value.init_offs = 0;
	attr_char_value.max_len = NRF_BLE_MAX_MTU_SIZE;
	memset(&Data, 0xff, sizeof(MPU_Data_t));
	attr_char_value.p_value = (uint8_t *)&Data; //???

    err_code = sd_ble_gatts_characteristic_add(m_service_handle,
                                               &char_md,
                                               &attr_char_value,
                                               &m_char_handles);

	print(LL_PRINT, "Characteristic handle = %d, CCCD handle = %d\n", m_char_handles.value_handle, m_char_handles.cccd_handle);

	APP_ERROR_CHECK(err_code);
}

/***********************************************************/ /**
 *@brief Function for adding the Service.
 *
 * @details This function adds the service and the characteristic within it to the local db.
 *
 **************************************************************/
void serviceAdd(void)
{
	ble_uuid_t service_uuid;
	uint32_t err_code;

	service_uuid.uuid = LOCAL_SERVICE_UUID;

	err_code = sd_ble_uuid_vs_add(&m_base_uuid128, &service_uuid.type);
	APP_ERROR_CHECK(err_code);

	err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &service_uuid, &m_service_handle);
	APP_ERROR_CHECK(err_code);

	// Add characteristics
	char_add(service_uuid.type);
}

/***********************************************************/ /**
 * @brief Function for starting advertising.
 **************************************************************/
void advertisingStart(void)
{
	uint32_t err_code;
	updateDeviceInfo();
	//printF("func : %s, line : %d\n",__func__,__LINE__);
	setMacAddr(globalData.macAddr);
	//printF("func : %s, line : %d\n",__func__,__LINE__);
	//err_code = sd_ble_gap_adv_start(&m_adv_params);
	err_code = sd_ble_gap_adv_start(&m_adv_params, APP_BLE_CONN_CFG_TAG);
	//printF("func : %s, line : %d\n",__func__,__LINE__);
	APP_ERROR_CHECK(err_code);
}

/***********************************************************/ /**
 *@brief Function for evaluating the value written in CCCD
 *
 * @details This shall be called when there is a write event received from the stack. This
 *          function will evaluate whether or not the peer has enabled notifications and
 *          start/stop notifications accordingly.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 **************************************************************/
static void on_write(ble_evt_t *p_ble_evt)
{
	ble_gatts_evt_write_t *p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;
	if ((p_evt_write->handle == m_char_handles.cccd_handle) && (p_evt_write->len == 2))
	{
		// Enable or disable notifications
		m_is_notifying_enabled = ble_srv_is_notification_enabled(p_evt_write->data);
	}
}

/***********************************************************/ /**
 *@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 **************************************************************/
static void bleGapEvtHandler(ble_evt_t *p_ble_evt)
{
	uint32_t err_code;

	switch (p_ble_evt->header.evt_id)
	{
	case BLE_GAP_EVT_CONNECTED:
		m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

		// Get the connection handle
		globalData.connectionHandle = p_ble_evt->evt.gap_evt.conn_handle;

		// Get the connection interval
		globalData.connectionInterval = p_ble_evt->evt.gap_evt.params.connected.conn_params.max_conn_interval;

		// Connection interval will be in units of 1.25 ms hence it should be converted into milliseconds and ceiled to next millisecond
		globalData.connectionInterval = globalData.connectionInterval * 125;												  // Multiply by 125
		globalData.connectionInterval = (globalData.connectionInterval / 100) + ((globalData.connectionInterval % 100) != 0); // Divide and ceil it to next millisecond

		break; // BLE_GAP_EVT_CONNECTED

	case BLE_GAP_EVT_DISCONNECTED:
		m_conn_handle = BLE_CONN_HANDLE_INVALID;

		// Reset the connection details
		globalData.connectionHandle = BLE_CONN_HANDLE_INVALID;
		globalData.connectionInterval = 0;
		globalData.bleDisconnectedFlag = true;
		//application_timers_stop();
		//Enque the BLE disconnect task.
		print(LL_DEBUG, "Host disconnected\n");
		taskEnqueue(globalData.bleDisconnectTaskId, NULL, 0, TASK_PRIORITY_HIGH, 0);

		break; // BLE_GAP_EVT_DISCONNECTED

	case BLE_GAP_EVT_TIMEOUT:

		if (p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_ADVERTISING)
		{
			// ??? to be handled
		}

		break; // BLE_GAP_EVT_TIMEOUT

	default:
		// No implementation needed.
		break;
	}
}

static void bleGattsEvtHandler(ble_evt_t *p_ble_evt)
{
	uint32_t err_code;
	uint8_t *mydata;

	switch (p_ble_evt->header.evt_id)
	{
	case BLE_GATTC_EVT_TIMEOUT:
		// Disconnect on GATT Client timeout event.
		err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
										 BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		APP_ERROR_CHECK(err_code);
		break; // BLE_GATTC_EVT_TIMEOUT

	case BLE_GATTS_EVT_TIMEOUT:
		// Disconnect on GATT Server timeout event.
		err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
										 BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
		APP_ERROR_CHECK(err_code);
		break; // BLE_GATTS_EVT_TIMEOUT

	case BLE_GATTS_EVT_WRITE:
		on_write(p_ble_evt);
		break; // BLE_GATTS_EVT_WRITE

	case BLE_EVT_USER_MEM_REQUEST:
		err_code = sd_ble_user_mem_reply(p_ble_evt->evt.gattc_evt.conn_handle, NULL);
		APP_ERROR_CHECK(err_code);
		break; // BLE_EVT_USER_MEM_REQUEST

	case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
	{
		ble_gatts_evt_rw_authorize_request_t req;
		ble_gatts_rw_authorize_reply_params_t auth_reply;

		req = p_ble_evt->evt.gatts_evt.params.authorize_request;

		if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
		{
			if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
			{
				auth_reply.params.write.gatt_status = bleAttributeWriteCallBack(
					p_ble_evt->evt.gatts_evt.params.authorize_request.request.write.uuid,
					p_ble_evt->evt.gatts_evt.params.authorize_request.request.write.handle,
					p_ble_evt->evt.gatts_evt.params.authorize_request.request.write.data,
					&(p_ble_evt->evt.gatts_evt.params.authorize_request.request.write.len));

				auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
				auth_reply.params.write.update = 1;
				auth_reply.params.write.offset = 0;
			}
			else
			{
				auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
				auth_reply.params.read.update = 1;
				auth_reply.params.read.offset = 0;

				auth_reply.params.read.p_data = (uint8_t *)readBuffer;
				auth_reply.params.read.len = 0;

				auth_reply.params.read.gatt_status = bleAttributeReadCallBack(
					p_ble_evt->evt.gatts_evt.params.authorize_request.request.read.uuid,
					p_ble_evt->evt.gatts_evt.params.authorize_request.request.read.handle,
					(uint8_t *)auth_reply.params.read.p_data,
					&(auth_reply.params.read.len));
			}

			err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
													   &auth_reply);
			APP_ERROR_CHECK(err_code);

			if ((m_conn_handle != BLE_CONN_HANDLE_INVALID)  
				&& ((globalData.upgradeTriggerFlag == true)
					|| (globalData.rebootTriggerFlag == true)
					|| (globalData.factoryResetTriggerFlag == true)
#if DEEP_SLEEP_ENABLED
					|| (globalData.deepSleepTriggerFlag == true)
#endif
					|| (globalData.resetPasswordWriteFlag == true)))
			{
				uint8_t data = NOTIFY_BEFORE_DISCONNECT;
				taskEnqueue(globalData.bleDisconnectTaskId, &data, sizeof(data), TASK_PRIORITY_HIGH, 0);
				//err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
				//m_conn_handle = BLE_CONN_HANDLE_INVALID;
				//print(LL_INFO, "error code = %d\n", err_code);
			}
		}
	}
	break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST

#if (NRF_SD_BLE_API_VERSION == 3)
	case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
		err_code = sd_ble_gatts_exchange_mtu_reply(p_ble_evt->evt.gatts_evt.conn_handle,
								NRF_BLE_MAX_MTU_SIZE);
		APP_ERROR_CHECK(err_code);
		break; // BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST
#endif

	default:
		// No implementation needed.
		break;
	}
}

/***********************************************************/ /**
 *@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the BLE Stack event interrupt handler after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 **************************************************************/
static void ble_evt_dispatch(ble_evt_t *p_ble_evt)
{
	if (p_ble_evt->header.evt_id == BLE_GAP_EVT_CONNECTED)
	{
		disconnectTimerInit();
	}

	disconnectTimerRestart();

	// Handle Gap events
	bleGapEvtHandler(p_ble_evt);

	// Handle GATTS events
	bleGattsEvtHandler(p_ble_evt);

	if (p_ble_evt->header.evt_id == BLE_GAP_EVT_DISCONNECTED)
	{
		disconnectTimerStop();
	}
}

/***********************************************************/ /**
 *@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 **************************************************************/
void sys_evt_dispatch(uint32_t sys_evt)
{
	// Beacon related handling
	appBeaconSdEvtSignalHandler(sys_evt);

	// Flash related handling
	fs_sys_event_handler(sys_evt);
}



/**@brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  Type of data to be looked for in advertisement data.
 * @param[in]  Advertisement report length and pointer to report.
 * @param[out] If data type requested is found in the data report, type data length and
 *             pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
    uint32_t  index = 0;
    uint8_t * p_data;

    p_data = p_advdata->p_data;

    while (index < p_advdata->data_len)
    {
        uint8_t field_length = p_data[index];
        uint8_t field_type   = p_data[index + 1];

        if (field_type == type)
        {
            p_typedata->p_data   = &p_data[index + 2];
            p_typedata->data_len = field_length - 1;
            return NRF_SUCCESS;
        }
        index += field_length + 1;
    }
    return NRF_ERROR_NOT_FOUND;
}


/**@brief Function for searching a given name in the advertisement packets.
 *
 * @details Use this function to parse received advertising data and to find a given
 * name in them either as 'complete_local_name' or as 'short_local_name'.
 *
 * @param[in]   p_adv_report   advertising data to parse.
 * @param[in]   name_to_find   name to search.
 * @return   true if the given name was found, false otherwise.
 */
static bool find_adv_name(ble_gap_evt_adv_report_t const * p_adv_report, char const * name_to_find)
{
    ret_code_t err_code;
    data_t     adv_data;
    data_t     dev_name;
    bool       found = false;

    // Initialize advertisement report for parsing.
    adv_data.p_data   = (uint8_t *)p_adv_report->data;
    adv_data.data_len = p_adv_report->dlen;

    // Search for matching advertising names.
    err_code = adv_report_parse(BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME, &adv_data, &dev_name);

    if (   (err_code == NRF_SUCCESS)
        && (strlen(name_to_find) == dev_name.data_len)
        && (memcmp(name_to_find, dev_name.p_data, dev_name.data_len) == 0))
    {
        found = true;
    }
    else
    {
        // Look for the short local name if the complete name was not found.
        err_code = adv_report_parse(BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME, &adv_data, &dev_name);

        if (   (err_code == NRF_SUCCESS)
            && (strlen(name_to_find) == dev_name.data_len)
            && (memcmp(m_target_periph_name, dev_name.p_data, dev_name.data_len) == 0))
        {
            found = true;
        }
    }
    return found;
}



/**@brief Function for handling BLE_GAP_ADV_REPORT events.
 * Search for a peer with matching device name.
 * If found, stop advertising and send a connection request to the peer.                                                                       
 */    
static void on_ble_gap_evt_adv_report(ble_gap_evt_t const * p_gap_evt)                                                                         
{
    if (!find_adv_name(&p_gap_evt->params.adv_report, m_target_periph_name))                                                                   
    {
        return;
    }  

    printf("Device \"%s\" found, sending a connection request.",
                 (uint32_t) m_target_periph_name);                                                                                             

    // Stop advertising.
    (void) sd_ble_gap_adv_stop();                                                                                                              

    // Initiate connection.
    m_conn_param.min_conn_interval = CONN_INTERVAL_DEFAULT;
    m_conn_param.max_conn_interval = CONN_INTERVAL_DEFAULT;                                                                                    

    ret_code_t err_code;
    err_code = sd_ble_gap_connect(&p_gap_evt->params.adv_report.peer_addr,
                                  &m_scan_param,                 
                                  &m_conn_param,                 
                                  APP_BLE_CONN_CFG_TAG);                                                                                       

    if (err_code != NRF_SUCCESS)   
    {  
        NRF_LOG_ERROR("sd_ble_gap_connect() failed: 0x%x.", err_code);                                                                         
    }  
}


/**@brief Function for handling BLE_GAP_EVT_CONNECTED events.
 * Save the connection handle and GAP role, then discover the peer DB.
 */
static void on_ble_gap_evt_connected(ble_gap_evt_t const * p_gap_evt)
{
    ret_code_t err_code;

    m_conn_handle = p_gap_evt->conn_handle;
    m_gap_role    = p_gap_evt->params.connected.role;

    if (m_gap_role == BLE_GAP_ROLE_PERIPH)
    {
        printf("Connected as a peripheral.\n");
    }
    else if (m_gap_role == BLE_GAP_ROLE_CENTRAL)
    {
        printf("Connected as a central.\n");
    }

    // Stop scanning and advertising.
    (void) sd_ble_gap_scan_stop();
    (void) sd_ble_gap_adv_stop();

//    bsp_board_leds_off();

    printf("Discovering GATT database...\n");
    err_code  = ble_db_discovery_start(&m_ble_db_discovery, p_gap_evt->conn_handle);
    printf("Discovering GATT database 111111111...\n");
    APP_ERROR_CHECK(err_code);

    if (m_gap_role == BLE_GAP_ROLE_PERIPH)
    {
    #if defined(S140)
        err_code = sd_ble_gap_phy_request(p_gap_evt->conn_handle, &m_test_params.phys);
        APP_ERROR_CHECK(err_code);
    #else
    printf(" m_gap_role == BLE_GAP_ROLE_PERIPH...\n");
        err_code = sd_ble_gap_phy_update(p_gap_evt->conn_handle, &m_test_params.phys);
        APP_ERROR_CHECK(err_code);
    #endif
    }
}


/**@brief Function for handling BLE_GAP_EVT_DISCONNECTED events.
 * Unset the connection handle and terminate the test.
 */
static void on_ble_gap_evt_disconnected(ble_gap_evt_t const * p_gap_evt)
{
    m_conn_handle = BLE_CONN_HANDLE_INVALID;

    printf("Disconnected: reason 0x%x.", p_gap_evt->params.disconnected.reason);

    if (m_run_test)
    {
        NRF_LOG_WARNING("GAP disconnection event received while test was running.");
    }

    bsp_board_leds_off();

    test_terminate();
}




/**@brief Function for handling BLE Stack events.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
    uint32_t              err_code;
    ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

	printf(" IN ble_evt_handler 1111==== %x\n", p_ble_evt->header.evt_id );
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_ADV_REPORT:
            on_ble_gap_evt_adv_report(p_gap_evt);
            break;

        case BLE_GAP_EVT_CONNECTED:
            on_ble_gap_evt_connected(p_gap_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_ble_gap_evt_disconnected(p_gap_evt);
            break;

        case BLE_GAP_EVT_CONN_PARAM_UPDATE:
        {
            m_conn_interval_configured = true;
            printf("Connection interval updated: 0x%x, 0x%x.",
                p_gap_evt->params.conn_param_update.conn_params.min_conn_interval,
                p_gap_evt->params.conn_param_update.conn_params.max_conn_interval);
        } break;

        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
        {
            // Accept parameters requested by the peer.
            ble_gap_conn_params_t params;
            params = p_gap_evt->params.conn_param_update_request.conn_params;
            err_code = sd_ble_gap_conn_param_update(p_gap_evt->conn_handle, &params);
            APP_ERROR_CHECK(err_code);

            printf("Connection interval updated (upon request): 0x%x, 0x%x.",
                p_gap_evt->params.conn_param_update_request.conn_params.min_conn_interval,
                p_gap_evt->params.conn_param_update_request.conn_params.max_conn_interval);
        } break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
        {
            err_code = sd_ble_gatts_sys_attr_set(p_gap_evt->conn_handle, NULL, 0, 0);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GATTC_EVT_TIMEOUT: // Fallthrough.
        case BLE_GATTS_EVT_TIMEOUT:
        {
            printf("GATT timeout, disconnecting.");
            err_code = sd_ble_gap_disconnect(m_conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_EVT_USER_MEM_REQUEST:
        {
            err_code = sd_ble_user_mem_reply(p_ble_evt->evt.common_evt.conn_handle, NULL);
            APP_ERROR_CHECK(err_code);
        } break;
        case BLE_GAP_EVT_PHY_UPDATE:
        {
            ble_gap_evt_phy_update_t const * p_phy_evt = &p_ble_evt->evt.gap_evt.params.phy_update;

            if (p_phy_evt->status == BLE_HCI_STATUS_CODE_LMP_ERROR_TRANSACTION_COLLISION)
            {
                // Ignore LL collisions.
                printf("LL transaction collision during PHY update.");
                break;
            }

            m_phy_updated = true;

            ble_gap_phys_t phys = {0};
            phys.tx_phys = p_phy_evt->tx_phy;
            phys.rx_phys = p_phy_evt->rx_phy;
            printf("PHY update %s. PHY set to %s.",
                         (p_phy_evt->status == BLE_HCI_STATUS_CODE_SUCCESS) ?
                         "accepted" : "rejected",
                         phy_str(phys));
        } break;

#if defined(S132)
        case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
        {
            err_code = sd_ble_gap_phy_update(p_gap_evt->conn_handle, &m_test_params.phys);
            APP_ERROR_CHECK(err_code);
        } break;
#endif

        default:
            // No implementation needed.
	printf(" IN ble_evt_handler 2222222222==== %x\n", p_ble_evt->header.evt_id );
			nrf_delay_ms(200);
            break;
    }
}


static void amtc_evt_handler(nrf_ble_amtc_t * p_amt_c, nrf_ble_amtc_evt_t * p_evt)
{
    ret_code_t err_code;

    switch (p_evt->evt_type)
    {
        case NRF_BLE_AMT_C_EVT_DISCOVERY_COMPLETE:
        {
            NRF_LOG_INFO("AMT service discovered at peer.");

            err_code = nrf_ble_amtc_handles_assign(p_amt_c,
                                                   p_evt->conn_handle,
                                                   &p_evt->params.peer_db);
            APP_ERROR_CHECK(err_code);

            // Enable notifications.
            err_code = nrf_ble_amtc_notif_enable(p_amt_c);
            APP_ERROR_CHECK(err_code);
        } break;

        case NRF_BLE_AMT_C_EVT_NOTIFICATION:
        {
            static uint32_t bytes_cnt  = 0;
            static uint32_t kbytes_cnt = 0;

            if (p_evt->params.hvx.bytes_sent == 0)
            {
                bytes_cnt  = 0;
                kbytes_cnt = 0;
            }

            bytes_cnt += p_evt->params.hvx.notif_len;
           if (bytes_cnt > 1024)
            {
                bsp_board_led_invert(PROGRESS_LED);

                bytes_cnt -= 1024;
                kbytes_cnt++;

                NRF_LOG_INFO("Received %u kbytes", kbytes_cnt);

                nrf_ble_amts_rbc_set(&m_amts, p_evt->params.hvx.bytes_rcvd);
            }

            if (p_evt->params.hvx.bytes_rcvd >= AMT_BYTE_TRANSFER_CNT)
            {
                bsp_board_led_off(PROGRESS_LED);

                bytes_cnt  = 0;
                kbytes_cnt = 0;

                NRF_LOG_INFO("Transfer complete, received %u bytes of ATT payload.",
                             p_evt->params.hvx.bytes_rcvd);

                nrf_ble_amts_rbc_set(&m_amts, p_evt->params.hvx.bytes_rcvd);
            }
        } break;

        case NRF_BLE_AMT_C_EVT_RBC_READ_RSP:
        {
            NRF_LOG_INFO("Peer received %u bytes of ATT payload.", (p_evt->params.rcv_bytes_cnt));
            test_terminate();
        } break;

        default:
            break;
    }
}


/**@brief Function for handling Database Discovery events.
 *
 * @details This function is a callback function to handle events from the database discovery module.
 *          Depending on the UUIDs that are discovered, this function should forward the events
 *          to their respective service instances.
 *
 * @param[in] p_evt  Pointer to the database discovery event.
 */
static void db_disc_evt_handler(ble_db_discovery_evt_t * p_evt)
{
    nrf_ble_amtc_on_db_disc_evt(&m_amtc, p_evt);
}



static void client_init(void)
{
	printf("client_init 11111111\n");
    ret_code_t err_code = ble_db_discovery_init(db_disc_evt_handler);
    APP_ERROR_CHECK(err_code);

	printf("client_init 22222222\n");
    err_code = nrf_ble_amtc_init(&m_amtc, amtc_evt_handler);
    APP_ERROR_CHECK(err_code);
}


static void server_init(void)
{
	printf("server_init 1111111111\n");
    nrf_ble_amts_init(&m_amts, amts_evt_handler);
}


/***********************************************************/ /**
 *@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 **************************************************************/
void bleStackInit(void)
{
#if 0
	uint32_t err_code=0;
	nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;

	// Initialize the SoftDevice handler module.
	SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    //ble_enable_params_t ble_enable_params; KARRY
    ble_cfg_t ble_enable_params;
    err_code = softdevice_enable_get_default_config(CENTRAL_LINK_COUNT,
                                                    PERIPHERAL_LINK_COUNT,
                                                    &ble_enable_params);
    APP_ERROR_CHECK(err_code);

	//Check the ram settings against the used number of links
	CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT, PERIPHERAL_LINK_COUNT);

// Enable BLE stack.
#if (NRF_SD_BLE_API_VERSION == 3)
	ble_enable_params.gatt_enable_params.att_mtu = NRF_BLE_MAX_MTU_SIZE;
#endif
	err_code = softdevice_enable(&ble_enable_params);
	APP_ERROR_CHECK(err_code);

	// Register with the SoftDevice handler module for BLE events.
	err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
	APP_ERROR_CHECK(err_code);

	// Register with the SoftDevice handler module for BLE events.
	err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
#else
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
	printf("--------bleStackInit --- ram_start %x\n", ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

	 printf("--------bleStackInit --- ram_start %x\n", ram_start);
	m_gap_role = BLE_GAP_ROLE_PERIPH; //KARRY

    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
	 printf("--------bleStackInit 2222222222--- ram_start %x\n", ram_start);
			nrf_delay_ms(200);

#endif
//APP_ERROR_CHECK(err_code);
}

/***********************************************************/ /**
 *@brief Function for the Power manager.
 **************************************************************/
void powerManage(void)
{
	uint32_t err_code = sd_app_evt_wait();
	APP_ERROR_CHECK(err_code);
}

/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module. This creates and starts application timers.
 */
static void timer_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
}

static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(app_timer_cnt_get);
    APP_ERROR_CHECK(err_code);
}

void cli_init(void)
{
    if (CoreDebug->DHCSR & CoreDebug_DHCSR_C_DEBUGEN_Msk)
    {
        ret_code_t err_code = nrf_cli_init(&m_cli_rtt, NULL, true, true, NRF_LOG_SEVERITY_INFO);
        APP_ERROR_CHECK(err_code);
    }

    nrf_drv_uart_config_t uart_config = NRF_DRV_UART_DEFAULT_CONFIG;
    uart_config.pseltxd = TX_PIN_NUMBER;
    uart_config.pselrxd = RX_PIN_NUMBER;
    uart_config.hwfc    = NRF_UART_HWFC_DISABLED;

    ret_code_t err_code = nrf_cli_init(&m_cli_uart, &uart_config, true, true, NRF_LOG_SEVERITY_INFO);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for the LEDs initialization.
 *
 * @details Initializes all LEDs used by the application.
 */
static void leds_init(void)
{
    bsp_board_leds_init();
}

/**@brief Function for enabling button input.
 */
static void buttons_enable(void)
{
    ret_code_t err_code = app_button_enable();
    APP_ERROR_CHECK(err_code);
}

static void buttons_disable(void)
{
    ret_code_t err_code = app_button_disable();
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling events from the button handler module.
 *
 * @param[in] pin_no        The pin that the event applies to.
 * @param[in] button_action The button action (press/release).
 */
static void button_evt_handler(uint8_t pin_no, uint8_t button_action)
{
    switch (pin_no)
    {
        case BOARD_TESTER_BUTTON:
        {
            printf("This board will act as tester.");
            instructions_print();
            m_board_role = BOARD_TESTER;
        } break;

        case BOARD_DUMMY_BUTTON:
        {
            printf("This board will act as responder.");
            m_board_role = BOARD_DUMMY;
            advertising_start();
            scan_start();
        } break;

        default:
            break;
    }
    buttons_disable();
}


void data_len_ext_set(bool status)
{
	printf("data_len_ext_set 11111111\n");
    m_test_params.data_len_ext_enabled = status;

    uint8_t data_length = status ? (247 + L2CAP_HDR_LEN) : (23 + L2CAP_HDR_LEN);
	printf("data_len_ext_set 222222222\n");
    (void) nrf_ble_gatt_data_length_set(&m_gatt, BLE_CONN_HANDLE_INVALID, data_length);
	printf("data_len_ext_set 3333333333\n");
}

static void conn_evt_len_ext_set(bool status)
{
    ret_code_t err_code;
    ble_opt_t  opt;

    memset(&opt, 0x00, sizeof(opt));
    opt.common_opt.conn_evt_ext.enable = status ? 1 : 0;

	printf("connection_len_ext_set 11111111\n");
    err_code = sd_ble_opt_set(BLE_COMMON_OPT_CONN_EVT_EXT, &opt);
	printf("connection_len_ext_set 222222222\n");
    APP_ERROR_CHECK(err_code);
}

void preferred_phy_set(ble_gap_phys_t * p_phy)
{
#if defined(S140)
    ble_opt_t opts;
    memset(&opts, 0x00, sizeof(ble_opt_t));
    memcpy(&opts.gap_opt.preferred_phys, p_phy, sizeof(ble_gap_phys_t));
    ret_code_t err_code = sd_ble_opt_set(BLE_GAP_OPT_PREFERRED_PHYS_SET, &opts);
    APP_ERROR_CHECK(err_code);
#endif
	printf("preferred_phy_set 11111111\n");
    memcpy(&m_test_params.phys, p_phy, sizeof(ble_gap_phys_t));
}


void gatt_mtu_set(uint16_t att_mtu)
{
    ret_code_t err_code;

	printf("gatt_mtu_set 11111111\n");
    m_test_params.att_mtu = att_mtu;
	printf("gatt_mtu_set 22222222\n");

    err_code = nrf_ble_gatt_att_mtu_periph_set(&m_gatt, att_mtu);
	printf("gatt_mtu_set 3333333\n");
    APP_ERROR_CHECK(err_code);

    err_code = nrf_ble_gatt_att_mtu_central_set(&m_gatt, att_mtu);
	printf("gatt_mtu_set 44444444\n");
    APP_ERROR_CHECK(err_code);
}




/**@brief Function for initializing the button library.
 */
static void buttons_init(void)
{
   // The array must be static because a pointer to it will be saved in the button library.
    static app_button_cfg_t buttons[] =
    {
        {BOARD_TESTER_BUTTON, false, BUTTON_PULL, button_evt_handler},
        {BOARD_DUMMY_BUTTON,  false, BUTTON_PULL, button_evt_handler}
    };

    ret_code_t err_code = app_button_init(buttons, ARRAY_SIZE(buttons), BUTTON_DETECTION_DELAY);
    APP_ERROR_CHECK(err_code);
}



/* Bluetooth initilizer */
void bleInit(void)
{
	printf("bleInit ---- 1111111\n");
	// BLE stack initialization
//	log_init();
#if 0	
	cli_init();
#endif
//	leds_init();
	timer_init();
	counter_init();
//	    buttons_init();
	bleStackInit(); //KARRY
	printf("bleInit ---- 222222\n");
	nrf_delay_ms(250);
	// GAP params initialization
	gapParamsInit();
	nrf_delay_ms(250);
	gatt_init();  //KARRY
	printf("bleInit ---- 333333\n");
	// Advertising initialization
	connectableAdvInit();
	printf("bleInit ---- 44444444\n");

//	advertising_start();
  //      scan_start();


	// Service Addition
	serviceAdd();
	printf("bleInit ---- 555555555\n");

	// Advertising data initialization
	advertisingDataInit();
	printf("bleInit ---- 6666666\n");

	//server_init();
    client_init();
	    gatt_mtu_set(m_test_params.att_mtu);
    data_len_ext_set(m_test_params.data_len_ext_enabled);
    conn_evt_len_ext_set(m_test_params.conn_evt_len_ext_enabled);
    preferred_phy_set(&m_test_params.phys);

	// Start the advertisement
	 advertisingStart();

        scan_start();
	// Delay before beacon initilization
	nrf_delay_ms(10);

	// Start beacon advertising
	appBeaconStart();
	printf("bleInit ---- 77777777\n");

	// Set the transmit power
	sd_ble_gap_tx_power_set(globalData.txPower);
	printf("bleInit ---- 888888888\n");
}
